"""
Created on 2022-08-22

@author: wf
"""
import justpy as jp
import pandas as pd

df_object = {"row1": [], "row2": [], "row3": [], "row4": [], "row5": []}
df_object["row1"] = [pd.Timestamp(year=2016, month=4, day=8)] * 5
df_object["row2"] = ["100"] * 5
df_object["row3"] = [100.345] * 5
df_object["row4"] = [3] * 5
df_object["row5"] = [100] * 5

df = pd.DataFrame(df_object)
msg_div = None


async def row_selected(self, msg):
    print(msg)
    if msg.selected:
        self.a.components[0].text = msg.data
        self.row_selected = msg.rowIndex
    elif self.row_selected == msg.rowIndex:
        self.row_data_div.text = ""


async def btn_click(self, msg):
    print("Button clicked")
    self.a.components[0].text = "Clicked!!!"


def grid_test_issue270():
    wp = jp.WebPage()
    row_data_div = jp.Div(a=wp, text="Replace me by selecting a row")
    grid = df.jp.ag_grid(a=wp)
    grid.on("rowSelected", row_selected)
    grid.options.columnDefs[0].checkboxSelection = True
    btn = jp.Button(text="Click", a=wp, click=btn_click)
    return wp


from examples.basedemo import Demo

Demo("Issue 270 aggrid timestamp value", grid_test_issue270)



# Justpy Tutorial demo html_comps4 from docs/tutorial/html_components.md
import justpy as jp

def html_comps4():
    wp = jp.WebPage()
    for i in range(10):
        d = jp.Div(a=wp, classes='m-2')
        for j in range(10):
            jp.Span(text=f'Span #{j+1} in Div #{i+1}', a=d, classes='text-white bg-blue-700 hover:bg-blue-200 ml-1 p-1')
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("html_comps4",html_comps4)



import justpy as jp
import pandas as pd
import numpy as np
import asyncio

# https://worldhappiness.report/ed/2019/
df = pd.read_csv("http://elimintz.github.io/happiness_report_2019.csv").round(3)


def grid_change(self, msg):
    msg.page.df = jp.read_csv_from_string(msg.data)
    c = msg.page.df.jp.plot(
        "Country",
        msg.page.cols_to_plot,
        kind="column",
        title="World Happiness Ranking",
        subtitle="Click and drag in the plot area to zoom in. Shift + drag to pan",
        stacking=msg.page.stacking,
        temp=True,
    )
    msg.page.c.options = c.options


def stack_change(self, msg):
    msg.page.c.options.plotOptions.series.stacking = self.value
    msg.page.stacking = self.value


def series_change(self, msg):
    msg.page.cols_to_plot = []
    for i, toggle in enumerate(msg.page.toggle_list):
        if toggle.value:
            msg.page.cols_to_plot.append(i + 3)
    c = msg.page.df.jp.plot(
        "Country",
        msg.page.cols_to_plot,
        kind="column",
        title="World Happiness Ranking",
        subtitle="Click and drag in the plot area to zoom in. Shift + drag to pan",
        stacking=msg.page.stacking,
        temp=True,
    )
    msg.page.c.options = c.options


async def corr_button_click(self, msg):
    msg.page.open = "/corr"
    await msg.page.update()
    msg.page.open = ""
    return True


def happiness_plot(request):
    # ['Country', 'Rank', 'Score', 'Unexplained', 'GDP', 'Social_support', 'Health', 'Freedom', 'Generosity', Corruption']
    wp = jp.QuasarPage()
    chart_theme = request.query_params.get("theme", 8)  # Default is 'grid'
    themes = [
        "high-contrast-dark",
        "high-contrast-light",
        "avocado",
        "dark-blue",
        "dark-green",
        "dark-unica",
        "gray",
        "grid-light",
        "grid",
        "sand-signika",
        "skies",
        "sunset",
    ]
    wp.highcharts_theme = themes[int(chart_theme)]
    wp.stacking = ""
    wp.cols_to_plot = [3, 4, 5, 6, 7, 8, 9]
    wp.df = df
    d = jp.Div(classes="q-ma-lg", a=wp)
    bg = jp.QBtnToggle(
        push=True,
        glossy=True,
        toggle_color="primary",
        value="",
        a=d,
        input=stack_change,
        style="margin-right: 30px",
    )
    bg.options = [
        {"label": "No Stacking", "value": ""},
        {"label": "Normal", "value": "normal"},
        {"label": "percent", "value": "percent"},
    ]
    wp.toggle_list = []
    corr_button = jp.QBtn(
        label="Show pairwise correlation", a=d, click=corr_button_click
    )
    for i in df.columns[3:]:
        wp.toggle_list.append(
            jp.QToggle(
                checked_icon="check",
                color="green",
                unchecked_icon="clear",
                value=True,
                label=f"{i}",
                a=d,
                input=series_change,
            )
        )  # ,style='margin-right: 10px',
    chart = df.jp.plot(
        "Country",
        wp.cols_to_plot,
        kind="column",
        a=wp,
        title="World Happiness Ranking",
        subtitle="Click and drag in the plot area to zoom in. Shift + drag to pan",
        stacking="",
        classes="border m-2 p-2 q-ma-lg p-ma-lg",
    )
    grid = df.jp.ag_grid(a=wp)
    grid.options.columnDefs[0].rowDrag = True
    for event_name in ["sortChanged", "filterChanged", "columnMoved", "rowDragEnd"]:
        grid.on(event_name, grid_change)
    wp.c = chart
    grid.c = chart
    bg.c = chart
    return wp


class ScatterWithRegression(jp.Scatter):
    def __init__(self, x, y, **kwargs):
        super().__init__(x, y, **kwargs)
        x = np.asarray(x)
        y = np.asarray(y)
        m = (len(x) * np.sum(x * y) - np.sum(x) * np.sum(y)) / (
            len(x) * np.sum(x * x) - np.sum(x) ** 2
        )
        b = (np.sum(y) - m * np.sum(x)) / len(x)
        s = jp.Dict()  # The new series
        s.type = "line"
        s.marker.enabled = False
        s.enableMouseTracking = False
        min = float(x.min())
        max = float(x.max())
        s.data = [[min, m * min + b], [max, m * max + b]]
        s.name = f"Regression, m: {round(m, 3)}, b: {round(b, 3)}"
        self.options.series.append(s)


chart_list = []


def create_corr_page():
    # creates a page showing correlation between factors
    score_factors = [
        "Unexplained",
        "GDP",
        "Social_support",
        "Health",
        "Freedom",
        "Generosity",
        "Corruption",
    ]
    wp = jp.WebPage(delete_flag=False)
    d = jp.Div(a=wp, classes="flex flex-wrap")
    for x_col in score_factors:
        for y_col in score_factors[score_factors.index(x_col) + 1 :]:
            x = df[x_col]
            y = df[y_col]
            chart = ScatterWithRegression(
                x,
                y,
                a=d,
                style="width: 400px; height: 400px; margin: 10px;",
                classes="border",
            )
            chart_list.append(chart)
            o = chart.options
            o.title.text = f"{x_col} vs. {y_col}"
            o.xAxis.title.text = x_col
            o.yAxis.title.text = y_col
            o.tooltip.pointFormat = f"{{point.country}}<br/></b>{x_col}: <b>{{point.x}}</b><br/></b>{y_col}: <b>{{point.y}}</b><br/>"
            o.tooltip.useHtml = True
            o.series[0].name = f"{x_col} vs. {y_col}"
            data = []
            # Make the first series data a dictionary so that the country name  will be available to the tooltip
            for i, point in enumerate(o.series[0].data):
                data.append({"x": point[0], "y": point[1], "country": df["Country"][i]})
            o.series[0].data = data
    corr_def = (
        df[
            [
                "Unexplained",
                "GDP",
                "Social_support",
                "Health",
                "Freedom",
                "Generosity",
                "Corruption",
            ]
        ]
        .corr()
        .round(3)
    )
    corr_def.insert(loc=0, column="Factors/Factors", value=corr_def.index)
    g = corr_def.jp.ag_grid(a=wp, style="width: 90%; height: 250px", classes="m-2")
    return wp


corr_page = create_corr_page()


@jp.SetRoute("/corr")
def corr_test():
    return corr_page


async def page_ready(self, msg):
    wp = msg.page
    for chart in chart_list:
        wp.d.add(chart)
        await self.update()
        await asyncio.sleep(0.5)


# Loads charts one by one to page, improving user experience
@jp.SetRoute("/corr_staggered")
def corr_stag_test():
    wp = jp.WebPage()
    wp.on("page_ready", page_ready)
    wp.d = jp.Div(a=wp, classes="flex flex-wrap")
    return wp


from examples.basedemo import Demo

Demo("happiness plot demo", happiness_plot)



"""
utilities for justpy
"""
import asyncio
import inspect
import os

def find_files(path: str, ext: str) -> list:
        """
        find files with the given extension in the given path

        Args:
            path(str): the path to start with
            ext(str): the extension to search for

        Returns:
            list: a list of files found
        """
        foundFiles = []
        for root, _dirs, files in os.walk(path, topdown=False):
            for name in files:
                if name.endswith(ext):
                    filepath = os.path.join(root, name)
                    foundFiles.append(filepath)
        return foundFiles

def print_request(request):
    # See https://www.starlette.io/routing/ for path_params
    # and https://github.com/encode/starlette/blob/master/starlette/routing.py line 82
    print(type(request._scope))
    d = dict(request._scope)
    print(d)
    d.pop("headers")
    print(d)
    fields = [
        "path",
        "method",
        "url",
        "headers",
        "query_params",
        "path_params",
        "client",
        "cookies",
        "state",
    ]
    print("*************************************")
    for field in fields:
        # print(field, request[field])
        try:
            print(field, request[field])
        except:
            print(field, getattr(request, field))
    print(request.url.path, request.url.port, request.url.scheme, dir(request.url))
    for i, j in request.query_params.items():
        print(i, j)
    print("URL related -------")
    for j in [
        "components",
        "fragment",
        "hostname",
        "is_secure",
        "netloc",
        "password",
        "path",
        "port",
        "query",
        "replace",
        "scheme",
        "username",
    ]:
        print(j, getattr(request.url, j))
    for j in getattr(request.url, "components"):
        print(j)
    print("*************************************")


def run_task(task):
    """
    Helper function to facilitate running a task in the async loop
    """
    loop = asyncio.get_event_loop()
    loop.create_task(task)


async def create_delayed_task(task, delay, loop):
    await asyncio.sleep(delay)
    loop.create_task(task)


def print_func_info(*args):
    # Calling function name
    print(inspect.stack()[1][3])
    for i in args:
        print(i)



import typing
from jpcore.justpy_app import JustpyApp

class SetRoute:
    """
    Justpy specific route annotation
    
    """

    def __init__(self, route, **kwargs):
        """
        Constructor

        Args:
            route(Route): the starlette route to set
            **kwargs: Arbitrary keyword arguments.
        """
        self.route = route
        self.kwargs = kwargs

    def __call__(self, wpfunc:typing.Callable, **_instance_kwargs):
        """
        Args:
            wpfunc(Callable): a WebPage returning function
            **_instance_kwargs: Arbitrary keyword arguments (ignored).

        """
        # Create a new route
        app=JustpyApp.app
        app.add_jproute(path=self.route, wpfunc=wpfunc, name=self.kwargs.get("name", None))
        return wpfunc



# Justpy Tutorial demo check_test from docs/tutorial/input.md
import justpy as jp

def check_test():
    wp = jp.WebPage(data={'checked': True})
    label = jp.Label(a=wp, classes='m-2 p-2 inline-block')
    c = jp.Input(type='checkbox', classes='m-2 p-2 form-checkbox', a=label, model=[wp, 'checked'])
    caption = jp.Span(text='Click to get stuff', a=label)

    in1 = jp.Input(model=[wp, 'checked'], a=wp, classes='border block m-2 p-2')
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("check_test",check_test)



"""
Created on 2022-09-17

@author: th
"""
import asyncio
import unittest

from selenium.common import TimeoutException
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By

import justpy as jp
from tests.base_selenium_test import BaseSeleniumTest


@unittest.skipIf(True, "A fix was proposed activate once applied")
class TestIssue150(BaseSeleniumTest):
    """
    testing issue 150
    https://github.com/justpy-org/justpy/issues/150
    """

    async def setUp(self):
        await BaseSeleniumTest.setUp(self,port=8124)

    async def tearDown(self):
        self.browser.close()
        await asyncio.sleep(self.server.sleep_time)
        await self.server.stop()

    def issue_page(self):
        """
        webpage containing the issue
        """
        wp = jp.QuasarPage()
        self.date_field = QInputDateTime(
                filled=True,
                style='width: 600px',
                a=wp,
                classes="q-pa-md",
                value='2022-05-01 18:44',
                name="date_field"
        )
        return wp

    async def test_issue_150(self):
        """
        tests the presence of issue 150
        When selecting a new date through the date picker an infinite event loop is started.
        Cause found and documented in the issue.
        """
        await self.server.start(self.issue_page)
        url = self.server.get_url("/")
        self.browser.get(url)
        driver = self.get_waiting_browser(self.browser)
        input_field = self.browser.find_elements(By.TAG_NAME, "i")[0]
        input_field.click()
        day_fields_locator = (By.CLASS_NAME, "q-date__calendar-item")
        driver.until(EC.element_to_be_clickable(day_fields_locator))
        day_fields = self.browser.find_elements(*day_fields_locator)
        day_fields[20].click()
        click_count_locator = (By.NAME, "count_date_changes")
        # check that the click was registered
        self.assertTrue(
                driver.until(
                        EC.text_to_be_present_in_element(
                                locator=click_count_locator,
                                text_="1")))
        # check that the infinite loop does not start → count_date_changes should not change its value
        self.assertRaises(
                TimeoutException,
                lambda: driver.until_not(
                        EC.text_to_be_present_in_element(locator=click_count_locator, text_="1")))

    @unittest.skipIf(True, "Only for manual testing")
    def test_manual(self):
        """
        start the server to test the issue manually
        """
        jp.justpy(self.issue_page, port=8500)


class QInputDateTime(jp.QInput):
    """
    component to test
    see https://github.com/justpy-org/justpy/issues/150
    """

    def __init__(self, **kwargs):
        """
        constructor
        """
        super().__init__(**kwargs)
        self.mask = '####-##-## ##:##'

        # date selection
        date_slot = jp.QIcon(name='event', classes='cursor-pointer')
        c2 = jp.QPopupProxy(transition_show='scale', transition_hide='scale', a=date_slot)
        locale = {
            'days': 'Neděle_Pondělí_Úterý_Středa_Čtvrtek_Pátek_Sobota'.split('_'),
            'daysShort': 'Ne_Po_Út_St_Čt_Pá_So'.split('_'),
            'months': 'Leden_Únor_Březen_Duben_Květen_Červen_Červenec_Srpen_Září_Říjen_Listopad_Prosinec'.split('_'),
            'monthsShort': 'Led_Úno_Bře_Dub_Kvě_Čvn_Čvc_Srp_Zář_Říj_Lis_Pro'.split('_')
        }
        self.date = jp.QDate(
                mask='YYYY-MM-DD HH:mm',
                name='date_popup',
                a=c2,
                locale=locale
        )
        self.date.parent = self
        self.date.value = self.value
        self.date.on('input', self.date_time_change)
        self.prepend_slot = date_slot

        # time selection
        time_slot = jp.QIcon(name='access_time', classes='cursor-pointer')
        self.c2 = jp.QPopupProxy(transition_show='scale', transition_hide='scale', a=time_slot)
        self.time = jp.QTime(mask='YYYY-MM-DD HH:mm', format24h=True, name='time', a=self.c2)
        self.time.parent = self
        self.time.value = self.value
        self.time.on('input', self.date_time_change)
        self.append_slot = time_slot

        self.on('change', self.input_change)

        # ------ For debugging the issue --------------
        # count number of changes through the date popup
        self.count_date_changes = jp.QDiv(
                a=self,
                text="0",
                value=0,
                name="count_date_changes"
        )
        # ---------------------------------------------

    @staticmethod
    async def date_time_change(self, msg):
        """
        update all date time values
        """
        # ------ For debugging the issue --------------
        self.parent.count_date_changes.value += 1
        self.parent.count_date_changes.text = self.parent.count_date_changes.value
        # ---------------------------------------------
        print(self.value)
        self.parent.value = self.value  # set input field value
        self.parent.date.value = self.value  # set date selection value
        self.parent.time.value = self.value  # set time selection value

    @staticmethod
    def input_change(self, msg):
        """
        update all date time values
        """
        # ------ For debugging the issue --------------
        self.count_date_changes.value += 1
        self.count_date_changes.text = self.count_date_changes.value
        # ---------------------------------------------
        self.date.value = self.value
        self.time.value = self.value


'''
Created on 2022-09-14

@author: wf
'''
from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter
import asyncio
import gc
import justpy as jp
import os
import sys
import pandas as pd
import psutil
import socket
import traceback
from jpcore.demostarter import Demostarter
from jpcore.tutorial import TutorialManager,Example
from jpcore.demoapp import JustpyDemoApp
from pygments.lexers import get_lexer_by_name
from pygments.formatters.html import HtmlFormatter
from pygments import highlight
from justpy import parse_html, QBtn
from pydevd_file_utils import setup_client_server_paths

class BaseWebPage():
    """
    Base Webpage
    """
    def __init__(self):
        """
        constructor
        """
        self.debug=True
        
    def showError(self,msg_html:str):
        """
        show the given error
        
        Args:
            msg_html(str): the html code to show as an error
        """
        self.errors.inner_html=msg_html
        
    def get_html_error(self,ex)->str:
        """
        get the html error message for the given exception
        """
        error_msg=str(ex)
        trace=""
        if self.debug:
            trace=traceback.format_exc()
        if self.debug:
            print(error_msg)
            print(trace)
        error_msg_html=f"❌{error_msg}<pre>{trace}</pre>"
        return error_msg_html
      
    def handleException(self,ex):
        '''
        handle the given exception
        
        Args:
            ex(Exception): the exception to handle
        '''
        error_msg_html=self.get_html_error(ex)
        self.showError(error_msg_html)
        
    async def onPageReady(self,_msg:dict):   
        """
        react on page_ready event
        """
        process = psutil.Process(os.getpid())
        gc.collect()
        mem_message=(f'memory: {process.memory_info().rss:,} bytes')
        self.title_status.inner_html=mem_message
        
    def setup(self,extra_html=""):
        """
        prepare basic page infrastructure
        """
        self.wp=jp.QuasarPage()
        self.wp.head_html = '<link rel="stylesheet" href="/templates/css/pygments.css">'
        # RGB 49 65 153
        color="bg-indigo-800"
        #  <q-btn glossy label="Tutorial" classes="q-mr-sm" name="tutorial-btn"></q-btn>
        self.bp = parse_html(
        f"""
<div class="q-pa-md">
  <q-layout view="HHH lpr Fff" container  style="height: 99vh" class="shadow-2 rounded-borders">
    <q-header reveal class="{color}">
        <q-toolbar name="toolbar">
            <q-btn flat round dense icon="menu"/>
            <q-toolbar-title>Justpy examples</q-toolbar-title>
            <span class="label bg-primary text-white" name="title-status"></span>
            <q-space />
        </q-toolbar>
    </q-header>
    %s
    <q-page-container style="overflow: hidden;">
        <q-page padding name="main_page" style="overflow: hidden;">
            <div class="row">
              <div class="col-sm-12 name="error-display">
              </div>
            </div  
        </q-page>
    </q-page-container>    
    <q-footer class="{color} text-white">
        <div class="text-white q-pa-xs" style="height: 30px;" name="footer"></div>
    </q-footer>
</div>
""" %   extra_html,
        a=self.wp,
    )
        self.main_page = self.bp.name_dict["main_page"]
        self.footer = self.bp.name_dict["footer"]
        self.toolbar=self.bp.name_dict["toolbar"]
        self.title_status=self.bp.name_dict["title-status"]
        #self.tutorial_btn = self.bp.name_dict["tutorial-btn"]
        #self.tutorial_btn.on("click",self.on_tutorial_btn_click)
        # put into html
        self.errors=self.bp.name_dict["error-display"]
        self.wp.on("page_ready", self.onPageReady)

class DemoDisplay(BaseWebPage):
    """
    display for a single demo
    """
    def __init__(self,demo:JustpyDemoApp,example:Example=None):
        """
        constructor
        
        Args:
            demo:JustpyDemoApp: the demo to be displayed
            example:Example: the tutorial example this might relate to
        """
        BaseWebPage.__init__(self)
        self.demo=demo
        self.example=example
        self.lexer=get_lexer_by_name("python")
        self.formatter=HtmlFormatter()
        self.syntax_highlighted_code=highlight(self.demo.source,self.lexer, self.formatter)    
        
    def web_page(self):
        """
        return my WebPage
        """
        self.setup()
        example_source=self.demo.example_source
        self.container = jp.QDiv(a=self.main_page)
        self.demo_description = jp.QDiv(a=self.container, classes="row")
        self.example_source_link = jp.Div(a=self.demo_description)
        self.example_source_link.inner_html=example_source.img_link
        self.demo_desc = jp.QDiv(a=self.demo_description, classes="centered")
        self.demo_desc.inner_html=self.demo.source_link 
        if self.example:
            # special handling for tutorial code
            pass
        self.sourceFrame=jp.QDiv(a=self.container, classes="row")
        # display code
        self.source_code_div=jp.QDiv(a=self.sourceFrame, classes="col-6")
        self.source_code_div.inner_html=self.syntax_highlighted_code
        # display preview
        self.preview_div = jp.QDiv(a=self.sourceFrame, classes="col-6")
        self.toggle_btn = jp.QBtnToggle(
                a=self.preview_div,
                classes="row q-ma-md",
                toggle_color='blue',
                push=True,
                glossy=True,
                input=self.toggle_preview,
                value='video',
        )
        self.toggle_btn.options.append({'label': "Video demo", 'value': "video"})
        self.toggle_btn.options.append({'label': "Live demo", 'value': "live"})
        self.preview_container = jp.QDiv(a=self.preview_div, classes="row")
        self._showVideoDemo()
        # demo status
        self._showDemoStatus()
        return self.wp

    def toggle_preview(self, _msg):
        """
        toggle between demo video and iframe preview
        """
        if self.toggle_btn.value == "live":
            self._showLiveDemo()
        else:
            self._showVideoDemo()

    def _showDemoStatus(self):
        """
        shows the status of the live demo
        """
        jp.QDiv(a=self.footer, text=f"Status{self.demo.status}")

    def _showVideoDemo(self):
        """
        Display a video demonstrating the demo
        """
        self.preview_container.delete_components()
        self.preview_video = jp.QImg(
                a=self.preview_container,
                src=self.demo.video_url,
                alt=self.demo.description,
                classes="row",
                style="max-width: 800px"
        )

    def _showLiveDemo(self):
        """
        Display an iframe showing the demo
        """
        self.preview_container.delete_components()
        jp.Iframe(
                a=self.preview_container,
                src=f"/{self.demo.wpfunc.__name__}",
                classes="col-grow",
                style="overflow:hidden; height:100vh; width:100%;"
        )

class DemoBrowser(BaseWebPage):
    """
    Browser for demos
    """
    def __init__(self,base_path:str=None,debug:bool=False):
        """
        constructor
        
        Args:
            base_path(str): the path to search examples in
            debug(bool): if True switchon debugging
        """
        BaseWebPage.__init__(self)
        self.demo_starter=Demostarter(base_path=base_path,debug=debug)
        self.tutorial_manager=TutorialManager()
        jp.app.add_jproute("/demo/{demo_name}",self.show_demo)
        self.mounted={}
        self.page_load_count=0
        
    def optionalDebug(self,args):   
        '''
        start the remote debugger if the arguments specify so
        
        Args:
            args(): The command line arguments
        '''
        if args.debugServer:
            import pydevd
            print (args.debugPathMapping,flush=True)
            if args.debugPathMapping:
                if len(args.debugPathMapping)==2:
                    remotePath=args.debugPathMapping[0] # path on the remote debugger side
                    localPath=args.debugPathMapping[1]  # path on the local machine where the code runs
                    MY_PATHS_FROM_ECLIPSE_TO_PYTHON = [
                        (remotePath, localPath),
                    ]
                    setup_client_server_paths(MY_PATHS_FROM_ECLIPSE_TO_PYTHON)
                    #os.environ["PATHS_FROM_ECLIPSE_TO_PYTHON"]='[["%s", "%s"]]' % (remotePath,localPath)
                    #print("trying to debug with PATHS_FROM_ECLIPSE_TO_PYTHON=%s" % os.environ["PATHS_FROM_ECLIPSE_TO_PYTHON"]);
         
            pydevd.settrace(args.debugServer, port=args.debugPort,stdoutToServer=True, stderrToServer=True)
            print("command line args are: %s" % str(sys.argv))
        
    async def onSizeColumnsToFit(self,_msg:dict):   
        """
        handle event for sizing the ag_grid columns to fit
        """
        try:
            await asyncio.sleep(0.3)
            if self.ag_grid:
                await self.ag_grid.run_api('sizeColumnsToFit()', self.wp)
        except Exception as ex:
            self.handleException(ex)
            
    async def onPageReady(self,_msg:dict):   
        """
        react on page_ready event
        """
        try:
            await self.onSizeColumnsToFit(_msg)
            await super().onPageReady(_msg)
            if self.page_load_count==0:
                for i,demo_name in enumerate(
                        ["stock_test2",
                         "plot_test4",
                         "plotly_test",
                         "women_majors3",
                         "blackjack",
                         "dogs"]
                    ):
                    demo=self.demo_starter.demos_by_name[demo_name]
                    await self.add_demo(demo,i+1)
                pass
            else:
                for i,demo in enumerate(self.mounted):
                    await self.add_demo_tolist(demo)
                    
            self.page_load_count+=1
            self.footer.text=f"{len(self.mounted)} apps mounted {self.page_load_count} page loads"
        except Exception as ex:
            self.handleException(ex)
            
    async def add_demo(self,demo:JustpyDemoApp,index:int):
        """
        mount and add demo to list
        """
        try:
            self.mount(demo,index)
        except BaseException as ex:
            demo.status=self.get_html_error(ex)
        await self.add_demo_tolist(demo,self.video_size)
    
    async def add_demo_tolist(self,demo:JustpyDemoApp,video_size:int=512):
        """
        add a demo (potentially with video) to the thumbnail list
        
        Args:
            demo(JustpyDemoApp): the demo app to add
            video_size(int): the size of the video (default: 512)
        """
        video_link=demo.video_link(video_size=video_size)
        list_item_html = f"""<q-item clickable v-ripple>
                            <q-item-section avatar>{demo.example_source.img_link}</q-item-section>
                            <q-item-section>
                                {video_link}
                            </q-item-section>
                            <q-item-section avatar> 
                              <q-btn class="gt-xs text-grey-8" size="12px" flat dense round icon="delete" name="delete"/>
                            </q-item-section>
                            <q-separator spaced ></q-separator>
                        </q-item>"""
        list_item = parse_html(list_item_html)
        delete_button=list_item.name_dict["delete"]
        delete_button.video_list=self.video_list
        delete_button.list_item=list_item
        list_item.name = demo.name
        list_item.video_url = demo.video_url
        self.video_list.add_component(list_item,0)
        def delete_list_item(btn, _msg):
            """
            delete a list item
            """
            btn.video_list.remove(btn.list_item)
            pass
        list_item.name_dict["delete"].on("click", delete_list_item)
        
    def web_page(self):
        """
        browser for justpy demos
        """
        extra_html="""  <q-drawer name="drawer" width=784 breakpoint=700 show-if-above elevated content-class="bg-white text-blue">
            <q-scroll-area class="fit">
                <div class="q-pa-sm">
                    <q-list name="thumbnail_list">
                    </q-list>
                </div>
            </q-scroll-area>
    </q-drawer>"""
        self.setup(extra_html)
        try:
            self.video_list=self.bp.name_dict["thumbnail_list"]
            self.mount_all_btn=QBtn(label="Mount all",a=self.toolbar,classes="q-mr-sm",click=self.on_mount_all_btn_click)
            
            self.video_size=512
            icon_size=32
            lod=self.demo_starter.as_list_of_dicts()
            for record in lod:
                index=record["#"]
                demo=self.demo_starter.demos[index-1]
                example=self.tutorial_manager.examples_by_name.get(demo.name,None)
                if example is not None:
                    demo.example_source=example.example_source
                    record["→"]=example.example_source.img_link
            df=pd.DataFrame(lod)
            style='height: 90vh; width: 99%; margin: 0.25rem; padding: 0.25rem;'
            grid_options = """{
              'rowHeight': %s,
            defaultColDef: {
                filter: true,
                sortable: true,
                resizable: true,
                cellStyle: {textAlign: 'left'},
                headerClass: 'font-bold'
            }
        }""" % icon_size
            self.ag_grid=df.jp.ag_grid(a=self.main_page,style=style,options=grid_options )
            self.ag_grid.html_columns = [1,2,3,4,5]
            self.ag_grid.on('rowSelected', self.row_selected)
            self.ag_grid.options.columnDefs[0].checkboxSelection = True
        except Exception as ex:
            self.handleException(ex)
        return self.wp
    
    async def row_selected(self,msg):
        """
        a grid row has been selected
        """
        if msg.selected:
            try:
                row_data = msg.data
                index = row_data["#"]
                demo=self.demo_starter.demos[index-1]
                await self.add_demo(demo,index)
            except Exception as ex:
                self.handleException(ex)
        
    def mount(self,demo,index:int):
        """
        mount the given demo 
        
        Args:
            demo(JustpyDemoApp): the demo to mount
            index(int): the index to mount the demo at
        """
        try:
            total=len(self.demo_starter.demos)
            print (f"mounting {index}/{total}:{demo}")
            demo.mount(jp.app)
            demo.try_it_url=f"/{demo.wpfunc.__name__}"
            demo.status="✅"
            self.mounted[demo.name]=demo
        except BaseException as ex:
            demo.status=self.get_html_error(ex)
              
    async def on_mount_all_btn_click(self,_msg):
        """
        mount all button has been clicked
        """
        total=len(self.demo_starter.demos)
        print(f"mount all has been clicked ... trying to mount {total} demos ")
        for i,demo in enumerate(self.demo_starter.demos):
            self.mount(demo,i+1)
        await self.wp.update()
            
    def show_demo(self,request):
        """
        show a demo by name
        """
        if "demo_name" in request.path_params:
            demo_name=request.path_params["demo_name"]
            if demo_name in self.demo_starter.demos_by_name:
                demo=self.demo_starter.demos_by_name[demo_name]
                example=self.tutorial_manager.examples_by_name.get(demo_name,None)
                demo_display=DemoDisplay(demo,example)
                return demo_display.web_page()    
            else:
                self.showError(f"example {demo_name} not found")
        self.showError("no example demo name specified")
        

def main(argv=None):  # IGNORE:C0111
    """main program."""

    if argv is None:
        argv = sys.argv
    try:
        program_name = os.path.basename(sys.argv[0])
        parser = ArgumentParser(
            description="Justpy Examples browser",
            formatter_class=RawDescriptionHelpFormatter,
        )
        parser.add_argument(
            "-d", "--debug", dest="debug", action="store_true", help="show debug info"
        )
        parser.add_argument(
            "--heroku", action="store_true", help="in heroku environment pick up port from env variable and set host to 0.0.0.0"
        )
        parser.add_argument(
            "-p",
            "--path",
            default=os.getcwd(),
            help="path to the examples (default: %(default)s)",
        )
        parser.add_argument('--debugServer',
                                     help="remote debug Server")
        parser.add_argument('--debugPort',type=int,
                                     help="remote debug Port",default=5678)
        parser.add_argument('--debugPathMapping',nargs='+',help="remote debug Server path mapping - needs two arguments 1st: remotePath 2nd: local Path")
   
        parser.add_argument("--host", default=socket.getfqdn())
        parser.add_argument("--port", type=int, default=8000)
        args = parser.parse_args(argv[1:])
        demo_browser=DemoBrowser(base_path=args.path,debug=args.debug)
        if args.heroku:
            args.port=os.environ["PORT"]
            args.host="0.0.0.0"
        demo_browser.optionalDebug(args)
        jp.justpy(demo_browser.web_page,host=args.host, port=args.port,PLOTLY=True,KATEX=True,VEGA=True)
    except Exception as e:
        indent = len(program_name) * " "
        sys.stderr.write(program_name + ": " + repr(e) + "\n")
        sys.stderr.write(indent + "  for help use --help")
        if args.debug:
            print(traceback.format_exc())
        return 2    

if __name__ == "__main__":
    sys.exit(main())




# Justpy Tutorial demo women_majors1 
# generated by demoapp.update_from tutorial at 2022-09-21T08:58:28.071932+00:00 
# see https://github.com/justpy-org/justpy/blob/master/examples/charts_tutorial/pandas/women_majors1.py
# 
# Using the Pandas Extension            
# see https://justpy.io/charts_tutorial/pandas#using-the-pandas-extension
import justpy as jp
import pandas as pd


wm = pd.read_csv('https://elimintz.github.io/women_majors.csv').round(2)
# Create list of majors which start under 20% women students
wm_under_20 = list(wm.loc[0, wm.loc[0] < 20].index)

def women_majors1():
    wp = jp.WebPage()
    wm.jp.plot(0, 
      	wm_under_20, 
      	kind='spline', 
     	a=wp, 
      	title='The gender gap is transitory - even for extreme cases',
      	subtitle='Percentage of Bachelors conferred to women form 1970 to 2011 in the US for extreme cases where the percentage was less than 20% in 1970',
      	classes='m-2 p-2 w-3/4')
    return wp

# initialize the demo
from examples.basedemo import Demo
Demo("women_majors1", women_majors1)



# Justpy Tutorial demo out_test from docs/tutorial/handling_events.md
import justpy as jp

def click_out(self, msg):
    self.text = 'click out'
    self.set_classes('text-blue-500')

def click_in(self, msg):
    self.text = 'click in'
    self.set_classes('text-red-500')

def out_test():
    wp = jp.WebPage()
    for i in range(4):
        d = jp.Div(text=f'{i}) Div', a=wp, classes='m-4 p-4 text-xl border w-32')
        d.on('click__out', click_out)
        d.on('click', click_in)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("out_test",out_test)



# Justpy Tutorial demo input_demo5 from docs/tutorial/input.md
import justpy as jp

class InputWithDiv(jp.Div):

    input_classes = "m-2 bg-gray-200 border-2 border-gray-200 rounded w-64 py-2 px-4 text-gray-700 focus:outline-none focus:bg-white focus:border-purple-500"

    @staticmethod
    def input_handler(self, msg):
        self.div.text = self.value

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.in1 = jp.Input(a=self, classes=self.input_classes, placeholder='Please type here', input=self.input_handler)
        self.in1.div = jp.Div(text='What you type will show up here', classes='m-2 p-2 h-32 text-xl border-2', a=self)


def input_demo5(request):
    wp = jp.WebPage()
    for i in range(10):
        InputWithDiv(a=wp)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("input_demo5",input_demo5)



'''
Created on 2022-09-16

@author: wf
'''
import os
import re
from jpcore.utilities import find_files
from jpcore.example import Example, ExampleSource

class TutorialManager:
    """
    justpy tutorial
    """
    
    def __init__(self,docs_dir:str=None):
        """
        constructor
        
        Args:
            docs_dir(str): the path from which to read the tutorial files
        """
        self.tutorials={}
        self.script_dir = os.path.dirname(__file__)
        if docs_dir is None:
            docs_dir = f"{os.path.dirname(self.script_dir)}/docs"
        self.docs_dir=docs_dir
        self.total_lines=0
        self.total_examples=0
        self.examples_by_name={}
        self.tutorial_files = find_files(self.docs_dir, ".md")
        for tutorial_file_path in self.tutorial_files:
            tutorial=Tutorial(docs_dir,tutorial_file_path)
            self.tutorials[tutorial.name]=tutorial
            self.total_lines+=len(tutorial.lines)
            self.total_examples+=len(tutorial.examples)
            for example in tutorial.examples.values():
                if example.name in self.examples_by_name:
                    duplicate_example=self.examples_by_name[example.name]
                    raise Exception(f"duplicate example name '{example.name}' {example.tutorial.name} and {duplicate_example.tutorial.name}")
                self.examples_by_name[example.name]=example
        
class Tutorial():
    """
    a single tutorial file
    """
    
    def __init__(self,docs_dir,path:str):
        """
        construct me
        """
        self.examples={}
        self.path=path
        self.name=path.replace(docs_dir+"/","")
        tutorial_path=self.name.replace(".md","")
        self.tutorial_url=f"https://justpy.io/{tutorial_path}"
        self.github_url=f"https://github.com/justpy-org/justpy/blob/master/docs/{self.name}"
        with open(self.path, "r") as markup_file:
            self.markup = markup_file.read()
        self.lines=self.markup.split("\n")
        header=None
        python_code=[]
        in_python_code=False
        for line in self.lines:
            if line.startswith("```python"):
                in_python_code=True
                python_code=[]
                continue
            if line.startswith("```"):
                in_python_code=False
                continue
            if not in_python_code:
                header_match= re.search("^#+\s*(.*)",line)
                if header_match:
                    header=header_match.group(1)
                    header=header.strip()
                    pass
            else:
                python_code.append(line)
                justpy_match = re.search(
                    """^(jp[.])?justpy[(]([a-zA-Z_0-9]*)([,]\s*(.*))?[)]([#].*)?""", line
                )
                if justpy_match:
                    example_name = justpy_match.group(2)
                    example_option=justpy_match.group(4)
                    example_comment=justpy_match.group(5)
                    if not example_name:
                        example_name=example_comment
                        pass
                    example_source=ExampleSource("tutorial")
                    example_source.lines=python_code
                    if header is not None:
                        # https://stackoverflow.com/questions/72536973/how-are-github-markdown-anchor-links-constructed
                        lower=header.strip().lower().replace(" ","-")
                        anchor=""
                        for c in lower:
                            if c.isalnum() or c in "-_":
                                anchor+=c
                        example_source.url=f"{self.tutorial_url}#{anchor}"
                        example_source.description=header
                    example=Example(example_source,name=example_name,option=example_option,header=header,lines=python_code)
                    example.tutorial=self
                    self.examples[example.name]=example
                    header=None
                    python_code=[]
                    in_python_code=False
                    
    def __str__(self):
        """
        return my text representation
        
        Returns:
            str: a text representation of me
        """
        text=self.path
        return text


# Justpy Tutorial demo quasar_splitter_test from docs/quasar_tutorial/quasar_components.md
import justpy as jp

lorem = 'Lorem ipsum dolor sit, amet consectetur adipisicing elit. Quis praesentium cumque magnam odio iure quidem, quod illum numquam possimus obcaecati commodi minima assumenda consectetur culpa fuga nulla ullam. In, libero.'

def quasar_splitter_test():
    wp = jp.QuasarPage()

    before = jp.Div(classes='q-pa-md')
    jp.Div(text='Before', classes='text-h4 q-mb-md', a=before)
    for i in range(20):
        jp.Div(text=f'{i}. {lorem}', a=before, classes='q-my-md')

    after = jp.Div(classes='q-pa-md')
    jp.Div(text='After', classes='text-h4 q-mb-md', a=after)
    for i in range(20):
        jp.Div(text=f'{i}. {lorem}', a=after, classes='q-my-md')

    s = jp.QSplitter(style='height: 400px', a=wp, classes='q-ma-lg')
    s.separator_class='bg-orange'
    s.separator_style='width: 3px'
    s.before_slot = before
    s.after_slot = after

    chip = jp.QChip(a=wp, classes='q-ma-lg')
    value_avatar = jp.QAvatar(text='50', color='red', text_color='white', a=chip)
    jp.Span(text='Splitter value', a=chip)

    s.value_avatar = value_avatar

    def splitter_input(self, msg):
        self.value_avatar.text = int(self.value)

    s.on('input', splitter_input)
    s.separator_slot = value_avatar

    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("quasar_splitter_test",quasar_splitter_test)



# Justpy Tutorial demo todo_app2 from docs/blog/reactivity.md
import justpy as jp


def todo_app2():
    todos = ["Buy groceries", "Learn JustPy", "Do errands"]
    wp = jp.WebPage(tailwind=False)
    ol = jp.Ol(a=wp)
    for todo in todos:
        jp.Li(text=todo, a=ol)
    task = jp.Input(placeholder="Enter task", a=wp)

    def add_task(self, msg):
        todos.append(task.value)
        task.value = ""
        ol.delete_components()
        for todo in todos:
            jp.Li(text=todo, a=ol)

    jp.Button(text="Add Task", a=wp, click=add_task, style="margin-left: 10px")

    return wp


# initialize the demo
from examples.basedemo import Demo

Demo("todo_app2", todo_app2)



'''
Created on 2022-09-12

@author: wf
'''
from addict import Dict
import asyncio
import inspect
from types import MethodType

class WebPage:
    """
    a web page
    """ 
    # TODO: Add page events online, beforeunload, resize
    instances = {}
    sockets = {}
    next_page_id = 0
    use_websockets = True
    delete_flag = True
    tailwind = True
    debug = False
    highcharts_theme = None
    # One of ['avocado', 'dark-blue', 'dark-green', 'dark-unica', 'gray',
    # 'grid-light', 'grid', 'high-contrast-dark', 'high-contrast-light', 'sand-signika', 'skies', 'sunset']
    allowed_events = [
        "click",
        "visibilitychange",
        "page_ready",
        "result_ready",
        "keydown",
        "keyup",
        "keypress",
    ]

    def __init__(self, **kwargs):
        self.page_id = WebPage.next_page_id
        WebPage.next_page_id += 1
        self.cache = None  # Set this attribute if you want to use the cache.
        self.use_cache = False  # Determines whether the page uses the cache or not
        self.template_file = "tailwind.html"
        self.title = "JustPy"
        self.display_url = None
        self.redirect = None
        self.open = None
        self.favicon = None
        self.components = []  # list of direct children components on page
        self.cookies = {}
        self.css = ""
        self.head_html = ""
        self.body_html = ""
        # If html attribute is not empty, sets html of page directly
        self.html = ""
        self.body_style = ""
        self.body_classes = ""
        self.reload_interval = None
        self.events = []
        self.dark = (
            False  # Set to True for Quasar dark mode (use for other dark modes also)
        )
        self.data = {}
        WebPage.instances[self.page_id] = self
        for k, v in kwargs.items():
            self.__setattr__(k, v)
            
    def __repr__(self):
        return f"{self.__class__.__name__}(page_id: {self.page_id}, number of components: {len(self.components)}, reload interval: {self.reload_interval})"

    def __len__(self):
        return len(self.components)

    def add_component(self, child, position=None):
        if position is None:
            self.components.append(child)
        else:
            self.components.insert(position, child)
        return self

    async def on_disconnect(self, websocket=None):
        if self.delete_flag:
            self.delete_components()
            self.remove_page()

    def remove_page(self):
        WebPage.instances.pop(self.page_id)

    def delete_components(self):
        for c in self.components:
            c.delete()
        self.components = []

    def add(self, *args):
        for component in args:
            self.add_component(component)
        return self

    def __add__(self, other):
        self.add_component(other)
        return self

    def __iadd__(self, other):
        self.add_component(other)
        return self

    def remove_component(self, component):
        try:
            self.components.remove(component)
        except:
            raise Exception("Component cannot be removed because it was not in Webpage")
        return self

    def remove(self, component):
        self.remove_component(component)

    def get_components(self):
        return self.components

    def last(self):
        return self.components[-1]

    def set_cookie(self, k, v):
        self.cookies[str(k)] = str(v)

    def delete_cookie(self, k):
        if k in self.cookies:
            del self.cookies[str(k)]

    async def run_javascript(self, javascript_string, *, request_id=None, send=True):
        try:
            websocket_dict = WebPage.sockets[self.page_id]
        except:
            return self
        dict_to_send = {
            "type": "run_javascript",
            "data": javascript_string,
            "request_id": request_id,
            "send": send,
        }
        await asyncio.gather(
            *[
                websocket.send_json(dict_to_send)
                for websocket in list(websocket_dict.values())
            ],
            return_exceptions=True,
        )
        return self

    async def reload(self):
        return await self.run_javascript("location.reload()")

    async def update_old(self, *, built_list=None):
        try:
            websocket_dict = WebPage.sockets[self.page_id]
        except:
            return self
        if not built_list:
            component_build = self.build_list()
        else:
            component_build = built_list
        for websocket in list(websocket_dict.values()):
            try:
                WebPage.loop.create_task(
                    websocket.send_json(
                        {
                            "type": "page_update",
                            "data": component_build,
                            "page_options": {
                                "display_url": self.display_url,
                                "title": self.title,
                                "redirect": self.redirect,
                                "open": self.open,
                                "favicon": self.favicon,
                            },
                        }
                    )
                )
            except:
                print("Problem with websocket in page update, ignoring")
        return self

    async def update(self, websocket=None):
        """
        update the Webpage

        Args:
            websocket(): The websocket to use (if any)
        """
        try:
            websocket_dict = WebPage.sockets[self.page_id]
        except:
            return self
        page_build = self.build_list()
        dict_to_send = {
            "type": "page_update",
            "data": page_build,
            "page_options": {
                "display_url": self.display_url,
                "title": self.title,
                "redirect": self.redirect,
                "open": self.open,
                "favicon": self.favicon,
            },
        }

        if websocket:
            WebPage.loop.create_task(websocket.send_json(dict_to_send))
        else:
            websockets = list(websocket_dict.values())
            # https://stackoverflow.com/questions/54987361/python-asyncio-handling-exceptions-in-gather-documentation-unclear
            _results = await asyncio.gather(
                *[websocket.send_json(dict_to_send) for websocket in websockets],
                return_exceptions=True,
            )
            pass
        return self

    async def delayed_update(self, delay):
        await asyncio.sleep(delay)
        return await self.update()

    def to_html(self, indent=0, indent_step=0, format=True):
        block_indent = " " * indent
        if format:
            ws = "\n"
        else:
            ws = ""
        s = f"{block_indent}<div>{ws}"
        for c in self.components:
            s = f"{s}{c.to_html(indent + indent_step, indent_step, format)}"
        s = f"{s}{block_indent}</div>{ws}"
        return s

    def react(self):
        pass

    def build_list(self):
        object_list = []
        self.react()
        for i, obj in enumerate(self.components):
            obj.react(self.data)
            d = obj.convert_object_to_dict()
            object_list.append(d)
        return object_list

    def on(self, event_type, func):
        """
        add an event of the given event_type  with the given function
        
        Args:
            event_type(str): the type of the event without on_prefix
            func(Callable): the function to call
            
        """
        if event_type in self.allowed_events:
            if inspect.ismethod(func):
                setattr(self, "on_" + event_type, func)
            else:
                setattr(self, "on_" + event_type, MethodType(func, self))
            if event_type not in self.events:
                self.events.append(event_type)
        else:
            raise Exception(f"No event of type {event_type} supported")

    async def run_event_function(
        self, event_type, event_data, create_namespace_flag=True
    ):
        event_function = getattr(self, "on_" + event_type)
        if create_namespace_flag:
            function_data = Dict(event_data)
        else:
            function_data = event_data
        if inspect.iscoroutinefunction(event_function):
            event_result = await event_function(function_data)
        else:
            event_result = event_function(function_data)
        return event_result

    def add_event(self, event):
        if event not in self.allowed_events:
            self.allowed_events.append(event)


# Justpy Tutorial demo model_demo2 from docs/tutorial/model_and_data.md
import justpy as jp

corner_classes = 'p-3 absolute bg-gray-200 '

class MyDiv(jp.Div):

    def model_update1(self):
        # model has the form [wp, 'text'] for example
        if self.model[0].data[self.model[1]]:
            self.text = str(self.model[0].data[self.model[1]])
        else:
            self.text = "Nothing typed yet"


def model_demo2():
    wp = jp.WebPage()
    d = jp.Div(classes='relative h-screen bg-gray-600', a=wp, data={'text': ''})
    for v_pos in ['top', 'bottom']:
        for h_pos in ['left', 'right']:
            corner_div = jp.Div(classes=corner_classes + f'{v_pos}-0 {h_pos}-0', a=d)
            jp.Div(text=f'{v_pos} {h_pos}', a=corner_div)
            MyDiv(text=f'typing will go here', a=corner_div, model=[d, 'text'])
    middle_input = jp.Input(text='middle', classes='absolute text-xl border-2 border-red-600',
                            placeholder='Type here', style='top: 50%; left: 40%', model=[d, 'text'], a=d)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("model_demo2",model_demo2)



# Justpy Tutorial demo text_colors from docs/tutorial/tailwind.md
import justpy as jp

def text_colors():
    wp = jp.WebPage()
    d = jp.Div(classes='flex flex-wrap m-2', a=wp)
    for color in jp.Tailwind.tw_dict['text_color']:
        jp.Div(text=color, classes=f'{color} font-mono p-1 text-lg bg-blue-100 hover:bg-red-500 w-48', a=d)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("text_colors",text_colors)



# Justpy Tutorial demo grid_test from docs/tutorial/custom_components.md
import justpy as jp
import pandas as pd

class LinkedChartGrid(jp.Div):

    def __init__(self, df, x, y, **kwargs):
        super().__init__(**kwargs)
        self.df = df
        self.x = x
        self.y = y
        self.kind = kwargs.get('kind', 'column')
        self.stacking = kwargs.get('stacking', '')
        self.title = kwargs.get('title', '')
        self.subtitle = kwargs.get('subtitle', '')
        self.set_classes('flex flex-col')
        self.chart = df.jp.plot(x, y, a=self, classes='m-2 p-2 border', kind=self.kind, stacking=self.stacking, title=self.title, subtitle=self.subtitle)
        self.grid = df.jp.ag_grid(a=self)
        self.grid.parent = self
        for event_name in ['sortChanged', 'filterChanged', 'columnMoved', 'rowDragEnd']:
            self.grid.on(event_name, self.grid_change)


    @staticmethod
    def grid_change(self, msg):
        self.parent.df = jp.read_csv_from_string(msg.data)
        c = self.parent.df.jp.plot(self.parent.x, self.parent.y, kind=self.parent.kind, title=self.parent.title,
                                   subtitle=self.parent.subtitle, stacking=self.parent.stacking)
        self.parent.chart.options = c.options



alcohol_df = pd.read_csv('https://raw.githubusercontent.com/fivethirtyeight/data/master/alcohol-consumption/drinks.csv', encoding="ISO-8859-1")
bad_drivers_df = pd.read_csv('https://raw.githubusercontent.com/fivethirtyeight/data/master/bad-drivers/bad-drivers.csv', encoding="ISO-8859-1")


def grid_test():
    wp = jp.WebPage()
    c = LinkedChartGrid(alcohol_df, 0, [1,2,3,4], kind='column', a=wp, classes='m-4 p-2 border',
                        stacking='normal', title='Alcohol Consumption per Country', subtitle='538 data')
    LinkedChartGrid(bad_drivers_df, 0, [1,2,3,4,5,6,7], kind='column', a=wp, classes='m-4 p-2 border-4', title='Bad Drivers per US State', subtitle='538 data')
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("grid_test",grid_test)



# Justpy Tutorial demo plot_test2 from docs/tutorial/matplotlib.md
import justpy as jp
import matplotlib.pyplot as plt

def plot_test2():
    wp = jp.WebPage()

    first_figure = plt.figure()
    plt.plot([0, 1, 4, 9], marker='*', markersize=20, markeredgecolor='red')
    plt.title('First Figure Showing Second')
    plt.xlabel('x data')
    plt.ylabel('y data')

    second_figure = plt.figure()
    plt.plot([0, 1, 4, 9], marker='+', markersize=20, markeredgecolor='red')
    plt.title('Second Figure Showing First')
    plt.xlabel('x data')
    plt.ylabel('y data')

    # Keyword method of setting figure
    jp.Matplotlib(figure=second_figure, a=wp)
    # Using method after creation
    c = jp.Matplotlib(a=wp)
    c.set_figure(first_figure)

    # If memory may be an issue, don't forget to close figures not in use anymore
    plt.close(first_figure)
    plt.close(second_figure)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("plot_test2",plot_test2)



# Justpy Tutorial demo for_example1 from docs/blog/vue_comparison.md
import justpy as jp

def for_example1():
    wp = jp.WebPage(tailwind=False)  
    items = [{'message': 'Foo'},
             {'message': 'Bar'}]
    ul = jp.Ul(a=wp)
    for item in items:
        jp.Li(text=f'{item["message"]}', a=ul)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("for_example1",for_example1)



# Justpy Tutorial demo button_counter_demo_c from docs/blog/vue_comparison.md
import justpy as jp

class ButtonCounter(jp.Button):

    def __init__(self, **kwargs):
        self.count = 0
        super().__init__(**kwargs)
        self.on('click', self.button_clicked)

    def button_clicked(self, msg):
        self.count += 1

    def react1(self, data):
        self.text = f'You clicked me {self.count} times.'


async def button_counter_demo_c():
    wp = jp.WebPage(tailwind=False)
    for i in range(5):
        ButtonCounter(a=wp, count=i, style='margin: 10px')
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("button_counter_demo_c",button_counter_demo_c)



# Justpy Tutorial demo tool_tip_demo2 from docs/charts_tutorial/tooltips.md
import justpy as jp

# https://www.highcharts.com/demo/line-basic
chart_def = """
{
 chart: {
        type: 'spline'
    },
    title: {
        text: ''
    },
    subtitle: {
        text: 'JustPy Tooltip Demo'
    },
    yAxis: {
        title: {
            text: 'Number of Employees'
        }
    },
    legend: {
        layout: 'vertical',
        align: 'right',
        verticalAlign: 'middle'
    },
    plotOptions: {
        series: {
            pointStart: 2010
        }
    },
    series: [{
        name: 'Installation',
        data: [43934, 52503, 57177, 69658, 97031, 119931, 137133, 154175]
    }, {
        name: 'Manufacturing',
        data: [24916, 24064, 29742, 29851, 32490, 30282, 38121, 40434]
    }, {
        name: 'Sales & Distribution',
        data: [11744, 17722, 16005, 19771, 20185, 24377, 32147, 39387]
    }, {
        name: 'Project Development',
        data: [null, null, 7988, 12169, 15112, 22452, 34400, 34227]
    }, {
        name: 'Other',
        data: [12908, 5948, 8105, 11248, 8989, 11816, 18274, 18111]
    }]
}
"""


def tool_tip_demo2():
    wp = jp.WebPage()
    jp.Div(text='Charts with default Highcharts tooltip', classes='m-2 p-2 text-xl bg-blue-500 text-white',a =wp)
    d = jp.Div(classes='flex flex-wrap ', a=wp)
    my_charts = []
    for i in range(3):
        my_charts.append(jp.HighCharts(a=d, classes='m-2 p-2 border', options=chart_def))
    my_charts[0].options.tooltip.shared = False
    my_charts[1].options.tooltip.shared = True
    my_charts[2].options.tooltip.split = True
    my_charts[0].options.title.text = 'Simple Tooltip'
    my_charts[1].options.title.text = 'Shared Tooltip'
    my_charts[2].options.title.text = 'Split Tooltip'

    jp.Div(text='Charts with user defined tooltip formatter functions',
           classes='m-2 p-2 text-xl bg-blue-500 text-white', a=wp)
    d = jp.Div(classes='flex flex-wrap', a=wp)
    my_charts = []
    for i in range(3):
        my_charts.append(jp.HighCharts(a=d, classes='m-2 p-2 border', options=chart_def))

    my_charts[0].options.tooltip.shared = False
    my_charts[1].options.tooltip.shared = True
    my_charts[2].options.tooltip.split = True
    my_charts[0].options.title.text = 'Simple Tooltip - Formatter Example'
    my_charts[1].options.title.text = 'Shared Tooltip - Formatter Example'
    my_charts[2].options.title.text = 'Split Tooltip - Formatter Example'

    my_charts[0].on('tooltip', simple_tooltip_formatter)
    my_charts[1].on('tooltip', shared_tooltip_formatter)
    my_charts[2].on('tooltip', split_tooltip_formatter)
    return wp


async def simple_tooltip_formatter(self, msg):
    tooltip_html = f"""
    <div class="text-red-500">
    <div><span style="color: {msg.color}">&#x25CF;&nbsp;</span>{msg.series_name}</div>
    <div>Year: {msg.x}</div>
    <div>Number of employees: {"{:,}".format(msg.y)}</div>
    </div>
    """
    return await self.tooltip_update(tooltip_html, msg.websocket)


async def shared_tooltip_formatter(self, msg):
    tooltip_div = jp.Div(classes="text-white bg-blue-400")
    jp.Span(text=f'Year: {msg.x}', classes='text-lg', a=tooltip_div)
    for point in msg.points:
        point_div = jp.Div(a=tooltip_div)
        jp.Span(text=f'&#x25CF; {point.series_name}', classes='bg-white', style=f'color: {point.color}', a=point_div)
        jp.Span(text=f'Number of employees: {"{:,}".format(point.y)}', a=point_div)
    return await self.tooltip_update(tooltip_div.to_html(), msg.websocket)


async def split_tooltip_formatter(self, msg):
    tooltip_array = [f'The x value is {msg.x}']
    for point in msg.points:
        point_div = jp.Div()
        jp.Span(text='&#x25CF;', classes='bg-white', style=f'color: {point.color}', a=point_div)
        jp.Span(text=f'{point.series_name}', a=point_div)
        jp.Span(text=f'Year: {point.x}', a=point_div)
        jp.Span(text=f'Number of employees: {"{:,}".format(point.y)}', a=point_div)
        tooltip_array.append(point_div.to_html())
    return await self.tooltip_update(tooltip_array, msg.websocket)


# initialize the demo
from  examples.basedemo import Demo
Demo ("tool_tip_demo2",tool_tip_demo2)



'''
Created on 2022-09-17

@author: wf
'''
from difflib import Differ
import datetime
import importlib
import os
import re
import typing
from jpcore.justpy_app import JustpyApp
from starlette.routing import request_response
from jpcore.example import ExampleSource

class JustpyDemoApp:
    """
    a justpy application
    """

    def __init__(self,examples_dir:str,pymodule_file:str, wpfunc: typing.Callable = None, debug:bool=False,**kwargs):
        """
        Args:
            examples_dir(str): the root directory for the examples
            pymodule_file(str): the python module path where the app resides
            wpfunc(Callable): the webpage function to call
            debug(bool): if True switch on debug mode
            **kwargs: further keyword arguments to pass to the webpage function
        """
        self.debug=debug
        self.examples_dir=examples_dir
        self.pymodule_file = pymodule_file
        self.example_source=ExampleSource.of_path(pymodule_file)
        self.source=self.example_source.source
        self.check_demo()
        source_file=pymodule_file
        self.source_file=source_file.replace(self.examples_dir+"/","")
        url=f"https://github.com/justpy-org/justpy/blob/master/examples/{self.source_file}"
        self.example_source.url=url
        self.source_link=f"""<a href="{url}" target="_blank">{self.source_file}</a>"""
        self.video_url=None
        self.try_it_url=None
        self.name=os.path.basename(self.source_file).replace(".py","")
        self.wpfunc = wpfunc
        self.debug=debug
        self.status=""
        self.kwargs = kwargs

    def check_demo(self):
        """
        Check whether this is a demo
        """
        self.is_demo = False
        if "Demo(" or "Demo (" in self.source:
            #if self.debug:
            #    print(f"found Demo( in {self.pymodule_file}")
            func_match = re.search(
                """Demo[ ]?[(]["'](.*)["'],\s*(.*?)(,.*)*[)]""", self.source
            )
            if func_match:
                #len_groups = len(func_match.groups())
                self.description = func_match.group(1)
                self.wpfunc_name = func_match.group(2)
                self.wpfunc_params=func_match.group(3)
                # jenkins could have /var/lib/jenkins/jobs/justpy/workspace/examples/charts_tutorial/pandas/women_majors2.py is not a demo
                module_match = re.search(
                    "(examples/.*)[.]py", self.pymodule_file
                )
                if module_match:
                    self.pymodule = module_match.group(1)
                    self.pymodule = self.pymodule.replace("/", ".")
                    self.is_demo = not "lambda" in self.wpfunc_name                   
                    if self.debug and self.is_demo:
                        print(f"found {self.wpfunc_name}:{self.description} in module {self.pymodule} {self.pymodule_file}")
                        print(f"with params {self.wpfunc_params}")
                return
            
    def video_link(self,target_url:str=None,alt:str=None,title:str=None,video_size=512):
        """
        get a video link for this demo 
        
        Args:
            target_url(str): the url the video link should refer to
            alt(str): the alt image name to display - default is derived from demo name
            title(str): the tooltip/title to show - default is derived from demo name
            video_size(int): the size of the video - default: 512
            
        """
        video_url=self.video_url
        if video_url is None: 
            video_url="https://user-images.githubusercontent.com/1336221/190849659-1f998e61-9cf7-4824-9767-6711d9d6ed56.jpg"
        if target_url is None:
            target_url=self.try_it_url if self.try_it_url is not None else self.video_url
        if alt is None:
            alt=f"animation for {self.name}"
        if title is None:
            title=f"animation for {self.name}"
        video_link=f"""<a href="{target_url}" target="_blank"><img src="{video_url}" alt="{alt}" style="width:{video_size}px;height:{video_size}px;"></a>"""
        return video_link
    
    def update_from_tutorial(self,tutorial_source:ExampleSource,target_path:str=None):
        """
        update me from the given tutorial_source
        
        Args:
            tutorial_source(ExampleSource): - the tutorial content to update me from 
            target_path(str): where to store the updated result - if None store at the original location
        """
        if target_path is None:
            target_path=self.pymodule_file
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        timestamp=now.isoformat()
        comment=f"""# Justpy Tutorial demo {self.name} 
# generated by demoapp.update_from tutorial at {timestamp} 
# see {self.example_source.url}
# 
# {tutorial_source.description}            
# see {tutorial_source.url}"""                
        with open(target_path, 'w') as code:
            print(comment, file=code)
            for line in tutorial_source.lines:
                if line.startswith("jp.justpy("):
                    args=line.replace("jp.justpy(","")
                    demo_start=f"""# initialize the demo
from examples.basedemo import Demo
Demo("{self.name}", {args}"""
                    print(demo_start,file=code)
                else:
                    print(line,file=code)
                            
    def same_as_tutorial(self,tutorial_source:ExampleSource,debug:bool=False):
        """
        check that my content is the same as the tutorial
        
        Args:
            tutorial_source(ExampleSource): - the tutorial content to update me from 
            debug(bool): if True show debug information
        """
        d=Differ()
        demo_source=self.example_source
        if debug:
            print(self)
            print(tutorial_source)
            print(demo_source)
        diff=d.compare(tutorial_source.lines,demo_source.lines)
        problems=[]
        for i,diff_line in enumerate(diff):
            if i==0:
                if not "# Justpy Tutorial demo" in diff_line:
                    problems.append("Justpy Tutorial demo comment in line 1 missing")
            elif diff_line.startswith("-"):
                if diff_line.startswith("- # generated by demoapp"):
                    pass
                elif diff_line.startswith("- jp.justpy("):
                    pass
                else:
                    problems.append(f"removed line {i+1}:{diff_line}")
            elif diff_line.startswith("+"):
                if diff_line.startswith("+ #"):
                    pass
                elif diff_line.startswith("+ # generated by demoapp"):
                    pass
                elif diff_line.startswith("+ from examples.basedemo import Demo"):
                    pass
                elif diff_line.startswith("+ Demo"):
                    pass
                elif diff_line.startswith("+ from  examples.basedemo import Demo"):
                    pass
                elif diff_line[2:].strip():
                    problems.append(f"added line {i+1}:{diff_line}")
            if debug:
                print(f"  {i+1:3}:{diff_line}")
        return problems
        

    def mount(self, app:JustpyApp, name:str=None):
        """
        mount me on the given app with the given name

        Args:
            app(JustpyApp): the app to mount me on
            name(str): the name to use for mounting - if None use the function name
        """
        demo_module = importlib.import_module(self.pymodule)
        self.wpfunc = getattr(demo_module, self.wpfunc_name)
        if self.wpfunc is None:
            raise Exception(f"can't start {self.pymodule_file} -wpfunc/endoint is None")
        # https://fastapi.tiangolo.com/advanced/sub-applications/
        endpoint=app.response(self.wpfunc)
        # wrap/cast the endpoint for starlettes routing
        # https://github.com/encode/starlette/issues/734
        mount_app=request_response(endpoint)
        if name is None:
            name=self.wpfunc.__name__
        if self.debug:
            print(f"mounting  {str(self)}")
        app.mount(f"/{name}",mount_app,name)

    def __str__(self):
        """
        return my string representation
        
        Returns:
            str: a text representation
        """
        text = f"{self.pymodule}({self.example_source}):{self.wpfunc_name}:{self.description}"
        return text



# Justpy Tutorial demo altair_chart_test from docs/tutorial/altair.md
import altair as alt
import justpy as jp


# Load a simple dataset as a pandas DataFrame
from vega_datasets import data
cars = data.cars()

car_chart = alt.Chart(cars).mark_point().encode(
    x='Horsepower',
    y='Miles_per_Gallon',
    color='Origin',
).properties(
    width=500,
    height=300
)


def altair_chart_test(request):
    wp = jp.WebPage(tailwind=True)
    wp.altair_chart = jp.AltairChart(chart=car_chart, a=wp, classes='m-8')
    return wp


def change_limit(self, msg):
    wp = msg.page
    wp.chart.chart = create_chart(self.value)


def create_chart(value):
    source = alt.sequence(start=0, stop=value, step=0.1, as_='x')
    return alt.Chart(source).mark_line().transform_calculate(
        sin='sin(datum.x)',
        cos='cos(datum.x)'
    ).transform_fold(
        ['sin', 'cos']
    ).encode(
        x='x:Q',
        y='value:Q',
        color='key:N'
    ).properties(
        width=500,
        height=300
    )


@jp.SetRoute('/line')
def line_chart(request):
    wp = jp.QuasarPage(tailwind=True)
    jp.QSlider(a=wp, classes='m-8 p-2', style='width: 500px',input=change_limit, label=True, label_always=True,
               value=10, min=1, max=50, snap=True, markers=True)
    wp.chart = jp.AltairChart(chart=create_chart(10), a=wp, classes='m-4 p-2')
    return wp


# initialize the demo
from  examples.basedemo import Demo
Demo ("altair_chart_test",altair_chart_test, VEGA=True)



"""
Created on 2022-10-21

@author: wangw37
"""
import justpy as jp
import pandas as pd

class TestIssue578():
    """ 
    test for Issue 578
    
    see https://github.com/justpy-org/justpy/issues/578
    """
    def __init__(self,data_frames:list):
        """
        constructor
        """
        if len(data_frames)<1:
            raise Exception("There needs to be at least one data frame")
        self.data_frames=data_frames
        self.data_frame_index=0
        
    def next_index(self):
        """
        calculate the next index
        """
        return (self.data_frame_index+1) % len(self.data_frames)
    
    def next_button_text(self):
        """
        calculate the button text for selecting the next index
        """
        n_index=self.next_index()
        return f"show dataframe {n_index}"
        
    def content(self):
        """
        the browser main web page
        """
        self.wp =jp.WebPage()
        self.button = jp.Button(a=self.wp, text=self.next_button_text(), classes='mt-1 w-fit px-4 py-2 rounded-md border border-gray-300 shadow-sm bg-green-200 text-red-700')
        self.div = jp.Div(a=self.wp)
        self.ag_grid=jp.AgGrid(a=self.div)
        self.load_dataframe()
        self.button.on('click', self.button_click)
        return self.wp
    
    def load_dataframe(self):
        """
        load the dataframe for the current index
        """
        self.ag_grid.load_pandas_frame(self.data_frames[self.data_frame_index])
    
    async def button_click(self, _msg):
        """
        react on the button click
        """
        self.data_frame_index=self.next_index()
        self.load_dataframe()
        self.button.text=self.next_button_text()

def grid_test_issue578():
    """
    setup the TestIssue class and 
    """
    data_frame = pd.DataFrame(data= {'col1': [1, 2], 'col2': [11, 22]})
    data_frame1 = pd.DataFrame(data= {'col1': [10, 20], 'col2': [11, 22], 'col3': [111, 222]})
    testIssue578=TestIssue578(data_frames=[data_frame,data_frame1])
    return testIssue578.content()

from examples.basedemo import Demo
Demo("Issue 578 new_aggrid", grid_test_issue578)



# Justpy Tutorial demo input_test5 from docs/quasar_tutorial/QInput.md
import justpy as jp

def input_test5(request):
    wp = jp.QuasarPage(data={'text': ''})
    c1 = jp.Div(classes='q-pa-md', a=wp)
    c2 = jp.Div(classes='q-gutter-md', style='max-width: 300px', a=c1)
    c3 = jp.QInput(label='Standard', a=c2, model=[wp, 'text'])
    c4 = jp.QInput(filled=True, label='Filled', a=c2, model=[wp, 'text'])
    c5 = jp.QInput(outlined=True, label='Outlined', a=c2, model=[wp, 'text'])
    c6 = jp.QInput(standout=True, label='Standout', a=c2, model=[wp, 'text'])
    c7 = jp.QInput(standout='bg-teal text-white', label='Custom standout', a=c2, model=[wp, 'text'])
    c8 = jp.QInput(borderless=True, label='Borderless', a=c2, model=[wp, 'text'])
    c9 = jp.QInput(rounded=True, filled=True, label='Rounded filled', a=c2, model=[wp, 'text'])
    c10 = jp.QInput(rounded=True, outlined=True, label='Rounded outlined', a=c2, model=[wp, 'text'])
    c11 = jp.QInput(rounded=True, standout=True, label='Rounded standout', a=c2, model=[wp, 'text'])
    c12 = jp.QInput(square=True, filled=True, label='Square filled', hint='This is a hint', a=c2, model=[wp, 'text'])
    c13 = jp.QInput(square=True, outlined=True, label='Square outlined', a=c2, model=[wp, 'text'])
    c14 = jp.QInput(square=True, standout=True, label='Square standout', a=c2, model=[wp, 'text'])
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("input_test5",input_test5)



# Justpy Tutorial demo sine_test from docs/charts_tutorial/creating_charts.md
import justpy as jp
import numpy as np


def sine_test():
    wp = jp.WebPage()
    chart = jp.HighCharts(a=wp, classes="border m-2 p-2 w-3/4")
    o = chart.options
    o.title.text = "Sines Galore"
    x = np.linspace(-np.pi, np.pi, 201)
    for frequency in range(1,11):
        y = np.sin(frequency * x)
        s = jp.Dict()
        s.name = f"F{frequency}"
        s.data = list(zip(x, y))
        o.series.append(s)
    return wp


# initialize the demo
from examples.basedemo import Demo
Demo("sine_test", sine_test)



# Justpy Tutorial demo target_test from docs/tutorial/handling_events.md
import justpy as jp

class ButtonDiv(jp.Div):

    def __init__(self, **kwargs):

        super().__init__(**kwargs)
        for i in range(1,6):
            b = jp.Button(text=f'Button {i}', a=self, classes=' m-2 p-2 border text-blue text-lg')
            b.num = i
            b.on('click', self.button_clicked)
        self.info_div = jp.Div(text='info will go here', classes='m-2 p-2 border', a=self)

    def button_clicked(self, msg):
        print(self)
        print(msg.target)
        self.info_div.text = f'Button {msg.target.num} was clicked'

def target_test():
    wp = jp.WebPage()
    ButtonDiv(a=wp)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("target_test",target_test)



# Justpy Tutorial demo radio_test2 from docs/tutorial/input.md
import justpy as jp


def radio_changed(self, msg):
    self.result_div.text = ''
    d = jp.Div(a=self.result_div, classes='m-2 p-2 border')
    for btn in self.btn_list:
        if btn.checked:
            jp.Span(text=f'{btn.value} is checked', a=d, classes='text-green-500 mr-6')
        else:
            jp.Span(text=f'{btn.value} is NOT checked', a=d, classes='text-red-500 mr-6')


def radio_test2():
    wp = jp.WebPage()
    genders = ['male', 'female', 'other']
    ages = [(0, 30), (31, 60), (61, 100)]

    outer_div = jp.Div(classes='border m-2 p-2 w-64', a=wp)
    # Create div to show radio button selection but don't add yet to page. It will be added at the end
    # It is created here so that it could be assigned to the radio button attribute result_div
    result_div = jp.Div(text='Click radio buttons to see results here', classes='m-2 p-2 text-xl')

    jp.P(a=outer_div, text='Please select your gender:')
    gender_list = []
    for gender in genders:
        label = jp.Label(classes='inline-block mb-1 p-1', a=outer_div)
        radio_btn = jp.Input(type='radio', name='gender', value=gender, a=label, btn_list=gender_list,
                             result_div=result_div, change=radio_changed)
        gender_list.append(radio_btn)
        jp.Span(classes='ml-1', a=label, text=gender.capitalize())

    jp.Div(a=outer_div, classes='m-2')  # Add spacing and line break

    jp.P(a=outer_div, text='Please select your age:')
    age_list = []
    for age in ages:
        label = jp.Label(classes='inline-block mb-1 p-1', a=outer_div)
        radio_btn = jp.Input(type='radio', name='age', value=age[0], a=label, btn_list=age_list,
                             result_div=result_div, change=radio_changed)
        age_list.append(radio_btn)
        jp.Span(classes='ml-1', a=label, text=f'{age[0]} - {age[1]}')
        jp.Br(a=outer_div)

    wp.add(result_div)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("radio_test2",radio_test2)



# Justpy Tutorial demo dog_pic1 from docs/tutorial/request_object.md
import justpy as jp

async def dog_pic1(request):
    wp = jp.WebPage()
    breed = request.query_params.get('breed', 'papillon' )
    r = await jp.get(f'https://dog.ceo/api/breed/{breed}/images/random')
    img_url = r['message']
    jp.Img(src=img_url, a=wp, classes='m-2 p-2')
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("dog_pic1",dog_pic1)



# Justpy Tutorial demo radio_test1 from docs/tutorial/input.md
import justpy as jp

def radio_test1():
    wp = jp.WebPage()
    genders = ['male', 'female', 'other']
    ages = [(0, 30), (31, 60), (61, 100)]

    outer_div = jp.Div(classes='border m-2 p-2 w-64', a=wp)

    jp.P(a=outer_div, text='Please select your gender:')
    for gender in genders:
        label = jp.Label(classes='inline-block mb-1 p-1', a=outer_div)
        radio_btn = jp.Input(type='radio', name='gender', value=gender, a=label)
        jp.Span(classes='ml-1', a=label, text=gender.capitalize())

    jp.Div(a=outer_div, classes='m-2')  # Add spacing and line break
    jp.P(a=outer_div, text='Please select your age:')
    for age in ages:
        label = jp.Label(classes='inline-block mb-1 p-1', a=outer_div)
        radio_btn = jp.Input(type='radio', name='age', value=age[0], a=label)
        jp.Span(classes='ml-1', a=label, text=f'{age[0]} - {age[1]}')
        jp.Br(a=outer_div)

    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("radio_test1",radio_test1)



# Justpy Tutorial demo html_comps2 from docs/tutorial/html_components.md
import justpy as jp

def html_comps2():
    wp = jp.WebPage()
    jp.get_tag('i', text='Text in Italic', a=wp)
    jp.get_tag('br',a=wp)
    jp.get_tag('br',a=wp)
    jp.get_tag('br',a=wp)
    jp.get_tag('br',a=wp)
    jp.get_tag('strong', text='Text in the Strong element', a=wp)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("html_comps2",html_comps2)



# Justpy Tutorial demo html_comps1 from docs/tutorial/html_components.md
import justpy as jp

def html_comps1():
    wp = jp.WebPage()
    jp.I(text='Text in Italic', a=wp)
    jp.Br(a=wp)
    jp.Strong(text='Text in the Strong element', a=wp)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("html_comps1",html_comps1)



# Justpy Tutorial demo input_test9 from docs/quasar_tutorial/QInput.md
import justpy as jp

def input_test9(request):
    wp = jp.QuasarPage(data={'text': ''})
    c1 = jp.Div(classes='q-pa-md', a=wp)
    c2 = jp.Div(classes='q-gutter-md', style='max-width: 300px', a=c1)
    jp.QInput(filled=True, label='Phone', mask='(###) ### - ####', hint="Mask: (###) ### - ####", a=c2)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("input_test9",input_test9)



# Justpy Tutorial demo calculator_test1 from docs/tutorial/custom_components.md
import justpy as jp
from justpy import Div, Input, Button, WebPage, justpy

class Calculator(Div):

    btn_classes = 'w-1/4 text-xl font-bold p-2 m-1 border bg-gray-200 hover:bg-gray-700 shadow'
    layout_text = [['7', '8', '9', '*'], ['4', '5', '6', '-'], ['1', '2', '3', '+'], ['C', '0', '.', '=']]

    def __init__(self, **kwargs):

        super().__init__(**kwargs)
        self.value = 0
        self.tape = Input(classes='block p-2 m-2  border text-right text-sm bg-gray-200', a=self, readonly=True, value=' ', style='width: 90%')
        self.result = Input(classes='block p-2 m-2 border text-2xl text-right', a=self, readonly=True, value='0', style='width: 90%')
        for line in self.__class__.layout_text:
            d = Div(classes='flex w-auto m-2', a=self)
            for b in line:
                b1 = Button(text=b, a=d, classes=self.__class__.btn_classes, click=self.calculator_click)
                b1.calc = self

    @staticmethod
    def calculator_click(self, msg):
        calc = self.calc
        if self.text == 'C':
            calc.result.value = '0'
            calc.tape.value = ' '
            calc.value = 0
        elif self.text == '=':
            calc.result.value = str(eval(calc.tape.value))
            calc.value = eval(calc.tape.value)
            calc.tape.value = calc.result.value
        else:
            if calc.tape.value[-1] in '*+-/' or self.text in '*+-/':
                calc.tape.value += ' ' + self.text
            else:
                calc.tape.value += self.text
            try:
                calc.result.value = str(eval(calc.tape.value))
                calc.value = eval(calc.tape.value)
            except:
                pass


def calculator_test1():
    wp = WebPage()
    for i in range(10):
        c = Calculator(a=wp, classes='m-1 border inline-block', style='width: 250px')
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("calculator_test1",calculator_test1)



# Justpy Tutorial demo dialog_test from docs/quasar_tutorial/quasar_components.md
import justpy as jp

alert_dialog_html = """
<div class="q-pa-md q-gutter-sm">
    <q-btn label="Alert" color="primary" name="alert_button" />
    <q-dialog name="alert_dialog" persistent>
      <q-card>
        <q-card-section>
          <div class="text-h6">Alert</div>
        </q-card-section>

        <q-card-section>
          Lorem ipsum dolor sit amet consectetur adipisicing elit. Rerum repellendus sit voluptate voluptas eveniet porro. Rerum blanditiis perferendis totam, ea at omnis vel numquam exercitationem aut, natus minima, porro labore.
        </q-card-section>

        <q-card-actions align="right">
          <q-btn flat label="OK" color="primary" v-close-popup />
        </q-card-actions>
      </q-card>
    </q-dialog>
</div>
"""

seamless_dialog_html = """
<div class="q-pa-md q-gutter-sm">
    <q-btn label="Open seamless Dialog" color="primary" name="seamless_button" />

    <q-dialog seamless position="bottom" name="seamless_dialog">
      <q-card style="width: 350px">
        <q-linear-progress :value="0.6" color="pink" />

        <q-card-section class="row items-center no-wrap">
          <div>
            <div class="text-weight-bold">The Walker</div>
            <div class="text-grey">Fitz & The Tantrums</div>
          </div>

          <q-space />

          <q-btn flat round icon="play_arrow" />
          <q-btn flat round icon="pause" />
          <q-btn flat round icon="close" v-close-popup />
        </q-card-section>
      </q-card>
    </q-dialog>
  </div>
"""


def open_dialog(self, msg):
    self.dialog.value = True

def dialog_test():
    wp = jp.QuasarPage()

    c = jp.parse_html(alert_dialog_html, a=wp)
    c.name_dict["alert_button"].dialog = c.name_dict["alert_dialog"]
    c.name_dict["alert_button"].on('click', open_dialog)

    c = jp.parse_html(seamless_dialog_html, a=wp)
    c.name_dict["seamless_button"].dialog = c.name_dict["seamless_dialog"]
    c.name_dict["seamless_button"].on('click', open_dialog)

    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("dialog_test",dialog_test)



# Justpy Tutorial demo comp_test from docs/tutorial/input.md
import justpy as jp


def change_color(self, msg):
    self.color_div.set_class(f'bg-{self.value}-600')


def comp_test():
    wp = jp.WebPage()
    colors = ['red', 'green', 'blue', 'pink', 'yellow', 'teal', 'purple']
    select = jp.Select(classes='w-32 text-xl m-4 p-2 bg-white  border rounded', a=wp, value='red',
                  change=change_color)
    for color in colors:
        select.add(jp.Option(value=color, text=color, classes=f'bg-{color}-600'))
    select.color_div = jp.Div(classes='bg-red-600 w-32 h-16 m-4',a=wp)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("comp_test",comp_test)



# Justpy Tutorial demo event_demo3 from docs/tutorial/handling_events.md
import justpy as jp

def my_click3(self, msg):
    print(msg)
    self.text = 'I was clicked'

def event_demo3():
    wp = jp.WebPage()
    wp.debug = True
    d = jp.Div(text='Not clicked yet', a=wp, classes='w-48 text-xl m-2 p-1 bg-blue-500 text-white')
    d.on('click', my_click3)
    d.additional_properties =['screenX', 'pageY','altKey','which','movementX','button', 'buttons']
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("event_demo3",event_demo3)



# Justpy Tutorial demo rating_test1 from docs/quasar_tutorial/QRating.md
import justpy as jp


def rating_test1():
    wp = jp.QuasarPage(data={"rating": 2})
    d = jp.Div(classes="q-pa-md", a=wp)
    rating_div = jp.Div(classes="q-gutter-y-md column", a=d)
    jp.QRating(size="1.5em", icon="thumb_up", a=rating_div, model=[wp, "rating"])
    jp.QRating(
        size="2em",
        icon="favorite_border",
        color="red-7",
        a=rating_div,
        model=[wp, "rating"],
    )
    jp.QRating(
        size="2.5em",
        icon="create",
        color="purple-4",
        a=rating_div,
        model=[wp, "rating"],
    )
    jp.QRating(
        size="3em", icon="pets", color="brown-5", a=rating_div, model=[wp, "rating"]
    )
    jp.QRating(
        size="4.5em",
        icon="star_border",
        color="green-5",
        a=rating_div,
        model=[wp, "rating"],
    )
    jp.QRating(
        size="5em",
        icon="star_border",
        icon_selected="star",
        color="grey",
        a=rating_div,
        model=[wp, "rating"],
        color_selected=[
            "light-green-3",
            "light-green-6",
            "green",
            "green-9",
            "green-10",
        ],
    )
    jp.QRating(
        size="5em",
        icon="star_border",
        icon_selected="star",
        color="green-5",
        a=rating_div,
        model=[wp, "rating"],
    )
    jp.QRating(
        size="3.5em",
        max=4,
        color="red-5",
        a=rating_div,
        model=[wp, "rating"],
        icon=[
            "sentiment_very_dissatisfied",
            "sentiment_dissatisfied",
            "sentiment_satisfied",
            "sentiment_very_satisfied",
        ],
    )
    return wp


# initialize the demo
from examples.basedemo import Demo

Demo("rating_test1", rating_test1)



# Justpy Tutorial demo input_test11 from docs/quasar_tutorial/quasar_components.md
import justpy as jp

def input_test11():
    wp = jp.QuasarPage()
    in1 = jp.QInput(label='Enter email', style='width: 150px; margin: 20px', a=wp, lazy_rules=False)
    in1.rules = ["val => val.length <= 3 || 'Please use maximum 3 characters'"]
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("input_test11",input_test11)



# Justpy Tutorial demo focus_test from docs/reference/htmlcomponent.md
import justpy as jp

# Try not using this event handler and see what happens
def my_blur(self, msg):
        self.set_focus = False

def focus_test():
    wp = jp.WebPage()
    in1 = jp.Input(classes=jp.Styles.input_classes, placeholder='Input 1', a=wp, blur=my_blur)
    in2 = jp.Input(classes=jp.Styles.input_classes, placeholder='Input 2', a=wp, blur=my_blur)
    in3 = jp.Input(classes=jp.Styles.input_classes, placeholder='Input 3', a=wp, blur=my_blur)
    in4 = jp.Input(classes=jp.Styles.input_classes, placeholder='Input 4', a=wp, blur=my_blur)

    # Set focus on third Input element
    in3.set_focus = True

    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("focus_test",focus_test)



# Justpy Tutorial demo custom_comp_test1 from docs/tutorial/custom_components.md
import justpy as jp

class PillButton(jp.Button):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_classes('bg-blue-500 hover:bg-blue-700 text-white font-bold py-2 px-4 rounded-full')


def custom_comp_test1():
    wp = jp.WebPage()
    for i in range(5):
        PillButton(text='Pill Button', click='self.text="I was clicked"', a=wp, classes='m-2')
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("custom_comp_test1",custom_comp_test1)



# Justpy Tutorial demo custom_comp_test5 from docs/tutorial/custom_components.md
import justpy as jp

class CalendarDate(jp.Div):

    def __init__(self, **kwargs):
        self.month = 'Jan'
        self.year = '2010'
        self.weekday = 'Sun'
        self.day = '1'
        self.color = 'red'
        super().__init__(**kwargs)

    def react(self, data):
        self.inner_html = f"""
                        <div class="w-24 rounded-t overflow-hidden bg-white text-center m-2 cursor-default">
                            <div class="bg-{data['color']}-500 text-white py-1">
                                {self.month}
                            </div>
                            <div class="pt-1 border-l border-r">
                                <span class="text-4xl font-bold">{self.day}</span>
                            </div>
                            <div class="pb-2 px-2 border-l border-r border-b rounded-b flex justify-between">
                                <span class="text-xs font-bold">{self.weekday}</span>
                                <span class="text-xs font-bold">{self.year}</span>
                            </div>
                        </div>
                        """


def change_color(self, msg):
    self.d.data["color"] = self.value

def custom_comp_test5():
    wp = jp.WebPage()
    year = 2019
    month = 'Feb'
    d = jp.Div(classes='flex flex-wrap', a=wp, data={'color': 'purple'})
    for day in range(1,11):
        CalendarDate(day=day, month=month, year=year, color='teal', a=d)
    d = jp.Div(classes='flex flex-wrap', a=wp, data={'color': 'red'})
    for day in range(5, 26):
        c = CalendarDate(day=day, month='Jul', year='2005', color='yellow', a=d)
        c.day += 1    # Notice how the date starts from 6 and not from 5.
    colors = ['red', 'green', 'blue', 'pink', 'yellow', 'teal', 'purple']
    s = jp.Select(classes='w-32 text-xl m-2 p-2 bg-white  border rounded-full', a=wp, value='red',
                  change=change_color)
    s.d = d
    for color in colors:
        s.add(jp.Option(value=color, text=color, classes=f'bg-{color}-600'))
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("custom_comp_test5",custom_comp_test5)



# Justpy Tutorial demo for_example2 from docs/blog/vue_comparison.md
import justpy as jp

def for_example2():
    wp = jp.WebPage(tailwind=False)
    ul = jp.Ul(a=wp)
    ul.items = [{'message': 'Foo'},
             {'message': 'Bar'}]
    ul.parent_message = 'Parent'
    for index, item in enumerate(ul.items):
        jp.Li(text=f'{ul.parent_message}-{index}-{item["message"]}', a=ul)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("for_example2",for_example2)



# Justpy Tutorial demo input_test10 from docs/quasar_tutorial/QInput.md
import justpy as jp
import re

email_regex = re.compile(r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)")

def input_change(self, msg):
    print(self.value)
    if re.match(email_regex,self.value):
        self.error = False
    else:
        self.error = True
        self.error_message = 'Enter valid email address'
        self.bottom_slots = True


def input_test10():
    wp = jp.QuasarPage()
    in1 = jp.QInput(label='Enter email', style='width: 150px; margin: 20px', a=wp, input=input_change)
    return wp


# initialize the demo
from  examples.basedemo import Demo
Demo ("input_test10",input_test10)



# This example is based on a streamlit app: https://github.com/streamlit/demo-uber-nyc-pickups
import justpy as jp

import pandas as pd
import numpy as np
import pydeck as pdk

try:
    import altair as alt

    _has_altair = True
except:
    _has_altair = False


DATE_TIME = "date/time"
DATA_URL = (
    "http://s3-us-west-2.amazonaws.com/streamlit-demo-data/uber-raw-data-sep14.csv.gz"
)


def load_data(nrows):
    data = pd.read_csv(DATA_URL, nrows=nrows)
    lowercase = lambda x: str(x).lower()
    data.rename(lowercase, axis="columns", inplace=True)
    data[DATE_TIME] = pd.to_datetime(data[DATE_TIME])
    return data


data = load_data(100000)
lowercase = lambda x: str(x).lower()
data.rename(lowercase, axis="columns", inplace=True)
data[DATE_TIME] = pd.to_datetime(data[DATE_TIME])


def map_deck(data, lat, lon, zoom):
    deck = pdk.Deck(
        # https://docs.mapbox.com/help/getting-started/access-tokens/
        api_keys={"mapbox": "Insert your mapbox api token here"},
        map_provider="mapbox",
        map_style="light",
        initial_view_state={
            "latitude": lat,
            "longitude": lon,
            "zoom": zoom,
            "pitch": 50,
        },
        layers=[
            pdk.Layer(
                "HexagonLayer",
                data=data,
                get_position=["lon", "lat"],
                radius=100,
                elevation_scale=4,
                elevation_range=[0, 1000],
                pickable=True,
                extruded=True,
            ),
        ],
    )
    return deck


airports = {
    "La Guardia": [40.7900, -73.8700],
    "JFK": [40.6650, -73.7821],
    "Newark": [40.7090, -74.1805],
}

zoom_level_airports = 12
zoom_level_main = 11
midpoint = (np.average(data["lat"]), np.average(data["lon"]))


def create_histogram(hour_selected):
    filtered = data[
        (data[DATE_TIME].dt.hour >= hour_selected)
        & (data[DATE_TIME].dt.hour < (hour_selected + 1))
    ]
    hist = np.histogram(filtered[DATE_TIME].dt.minute, bins=60, range=(0, 60))[0]
    chart_data = pd.DataFrame({"minute": range(60), "pickups": hist})
    chart = (
        alt.Chart(chart_data)
        .mark_area(
            interpolate="step-after",
        )
        .encode(
            x=alt.X("minute:Q", scale=alt.Scale(nice=False)),
            y=alt.Y("pickups:Q"),
            tooltip=["minute", "pickups"],
        )
        .configure_mark(opacity=0.5, color="red")
        .properties(width="container", height=200)
    )
    return chart


async def slider_change(self, msg):
    wp = msg.page
    hour = int(self.value)
    wp.all_caption.text = f"All New York City from {hour}:00 to {hour+1}:00"
    hour_data = data[data[DATE_TIME].dt.hour == hour]
    wp.decks["main"].deck = map_deck(
        hour_data, midpoint[0], midpoint[1], zoom_level_main
    )
    for airport, coords in airports.items():
        wp.decks[airport].deck = map_deck(
            hour_data, coords[0], coords[1], zoom_level_airports
        )
    if _has_altair:
        wp.histogram.chart = create_histogram(self.value)
        wp.histogram_caption.text = (
            f"Breakdown of rides per minute between {hour}:00 and {hour+1}:00"
        )


def pydeck_demo(request):
    wp = jp.QuasarPage(tailwind=True, title="Uber NYC Pickups")
    wp.decks = {}
    slider_div = jp.Div(a=wp, classes="flex ml-4", style="width: 50%; margin-top: 20px")
    jp.Div(
        text="Select hour of pickup:", a=slider_div, classes="pt-3 text-xl font-bold "
    )
    s1 = jp.Span(classes="ml-6 mt-1", style="width: 50%", a=slider_div)
    jp.QSlider(
        classes="w-64",
        min=0,
        max=23,
        a=s1,
        label=True,
        label_always=True,
        markers=True,
        step=1,
        snap=True,
        color="red",
        change=slider_change,
    )
    deck_div = jp.Div(classes="flex ml-2", a=wp)
    hour_data = data[data[DATE_TIME].dt.hour == int(0)]
    deck = map_deck(hour_data, midpoint[0], midpoint[1], zoom_level_main)
    all_div = jp.Div(a=deck_div)
    wp.all_caption = jp.Div(
        text="All New York City from 0:00 to 1:00",
        a=all_div,
        style="margin: 10px;",
        classes="text-xl font-bold",
    )
    wp.decks["main"] = jp.PyDeckFrame(
        a=all_div,
        deck=deck,
        style="margin: 10px; height: 400px; width: 450px",
        transition_duration=0.5,
    )
    for airport, coords in airports.items():
        airport_deck = map_deck(hour_data, coords[0], coords[1], zoom_level_airports)
        airport_div = jp.Div(a=deck_div)
        jp.Div(
            text=f"{airport} Airport",
            a=airport_div,
            style="margin: 10px;",
            classes="text-xl font-bold",
        )
        wp.decks[airport] = jp.PyDeckFrame(
            a=airport_div,
            deck=airport_deck,
            style="margin: 10px; height: 400px; width: 250px",
            transition_duration=0.5,
        )
    if _has_altair:
        chart = create_histogram(0)
        wp.histogram_caption = jp.Div(
            text="Breakdown of rides per minute between 0:00 and 1:00",
            a=wp,
            style="margin-left: 30px;",
            classes="text-xl font-bold",
        )
        wp.histogram = jp.AltairChart(
            chart=chart, a=wp, style="padding: 10px; width: 100%;"
        )
    return wp


from examples.basedemo import Demo

Demo("uber pydeck demo", pydeck_demo, VEGA=True)



# Justpy Tutorial demo hello_world4 from docs/tutorial/basic_concepts.md
import justpy as jp

wp = jp.WebPage(delete_flag=False)
my_paragraph_design = (
    "w-64 bg-blue-500 m-2 hover:bg-blue-700 text-white font-bold py-2 px-4 rounded"
)
for i in range(1, 11):
    jp.P(text=f"{i}) Hello World!", a=wp, classes=my_paragraph_design)


def hello_world4():
    return wp


# initialize the demo
from examples.basedemo import Demo

Demo("hello_world4", hello_world4)



# Justpy Tutorial demo todo_app from docs/blog/vue_comparison.md
import justpy as jp


class TodoList(jp.Div):

    def __init__(self, **kwargs):
        self.todos = []
        super().__init__(**kwargs)
        self.ol = jp.Ol(a=self)
        self.task = jp.Input(placeholder='Enter task', a=self)
        jp.Button(text='Add Task', a=self, click=self.add_task, style='margin-left: 10px')

    def add_task(self, msg):
        self.todos.append(self.task.value)
        self.task.value = ''

    def react3(self, data):
        self.ol.delete_components()
        for todo in self.todos:
            jp.Li(text=todo, a=self.ol)


def todo_app():
    wp = jp.WebPage(tailwind=False)
    TodoList(a=wp, todos=['Buy groceries', 'Learn JustPy', 'Do errands'])
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("todo_app",todo_app)



# Justpy Tutorial demo count_test from docs/tutorial/pushing_data.md
import justpy as jp
import asyncio

button_classes = 'm-2 p-2 text-red-700 bg-white hover:bg-red-200 hover:text-red-500 border focus:border-red-500 focus:outline-none'

async def count_down(self, msg):
    self.show = False
    if hasattr(msg.page, 'd'):
        msg.page.remove(msg.page.d)
    bomb_icon = jp.Icon(icon='bomb', classes='inline-block text-5xl ml-1 mt-1', a=msg.page)
    d = jp.Div(classes='text-center m-4 p-4 text-6xl text-red-600 bg-blue-500 faster', a=msg.page, animation=self.count_animation)
    msg.page.d = d
    for i in range(self.start_num, 0 , -1):
        d.text = str(i)
        await msg.page.update()
        await asyncio.sleep(1)
    d.text = 'Boom!'
    d.animation = 'zoomIn'
    d.set_classes('text-red-500 bg-white')
    bomb_icon.set_class('text-red-700')
    self.show = True

def count_test(request):
    start_num = int(request.query_params.get('num', 10))
    animation = request.query_params.get('animation', 'flip')
    wp = jp.WebPage()
    count_button = jp.Button(text='Start Countdown', classes=button_classes, a=wp, click=count_down)
    count_button.start_num = start_num
    count_button.count_animation = animation
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("count_test",count_test)



# Justpy Tutorial demo plot_test4 from docs/tutorial/matplotlib.md
import justpy as jp
import matplotlib.pyplot as plt


def plot_test4():
    wp = jp.WebPage()
    print(f'start {plt.get_fignums()}')
    f = plt.figure()
    plt.plot([0, 1, 4, 9], marker='*', markersize=20, markeredgecolor='red')
    plt.title('Matplotlib Example')
    plt.xlabel('x data')
    plt.ylabel('y data')
    chart = jp.Matplotlib(a=wp)
    chart.num_points = 4
    plt.close(f)

    b = jp.Button(text='Add Point', a=wp,
                  classes='m-2 bg-transparent hover:bg-blue-500 text-blue-700 font-semibold hover:text-white py-2 px-4 border border-blue-500 hover:border-transparent rounded')
    b.chart = chart

    def add_point(self, msg):
        self.chart.num_points += 1
        np = self.chart.num_points
        f = plt.figure()
        plt.plot([i*i for i in range(np)], marker='*', markersize=20, markeredgecolor='red')
        plt.title(f'This chart has {np} points')
        plt.xlabel('x data')
        plt.ylabel('y data')
        self.chart.set_figure(f)
        plt.close(f)

    b.on('click', add_point)

    def subtract_point(self, msg):
        self.num_points -= 1
        np = self.num_points
        f = plt.figure()
        plt.plot([i*i for i in range(np)], marker='*', markersize=20, markeredgecolor='red')
        plt.title(f'This chart has {np} points')
        plt.xlabel('x data')
        plt.ylabel('y data')
        self.set_figure(f)
        plt.close(f)

    chart.on('click', subtract_point)

    return wp


# initialize the demo
from  examples.basedemo import Demo
Demo ("plot_test4",plot_test4)



# Justpy Tutorial demo message_demo from docs/tutorial/pushing_data.md

import justpy as jp
from datetime import datetime

input_classes = 'm-2 bg-gray-200 font-mono appearance-none border-2 border-gray-200 rounded py-2 px-4 text-gray-700 focus:outline-none focus:bg-white focus:border-red-500'
button_classes = 'm-2 p-2 text-red-700 bg-white hover:bg-red-200 hover:text-red-500 border focus:border-red-500 focus:outline-none'
message_classes = 'ml-4 p-2 text-lg bg-red-500 text-white overflow-auto font-mono rounded-lg'

shared_div = jp.Div(classes='m-2 h-1/2 border overflow-auto', delete_flag=False)
header = jp.Div(text='Simple Message Board', classes='bg-red-100 border-l-4 border-red-500 text-red-700 p-4 text-3xl', delete_flag=False)
button_icon = jp.Icon(icon='paper-plane', classes='text-2xl', delete_flag=False)
button_text = jp.Div(text='Send', classes='text-sm', delete_flag=False)
message_icon = jp.Icon(icon='comments', classes='text-2xl text-green-600', delete_flag=False)

def message_initialize():
    # Called once on startup
    d = jp.Div(a=shared_div, classes='flex m-2 border')
    time_stamp = jp.P(text=datetime.now().strftime('%Y-%m-%d %H:%M:%S'), classes='text-xs ml-2 flex-shrink-0')
    p = jp.Pre(text='Welcome to the simple message board!', classes=message_classes)
    d.add(message_icon, time_stamp, p)

async def send_message(self, msg):
    if self.message.value:
        d = jp.Div(classes='flex m-2 border')
        time_stamp = jp.P(text=datetime.now().strftime('%Y-%m-%d %H:%M:%S'), classes='text-xs ml-2 flex-shrink-0')
        p = jp.Pre(text=self.message.value, classes=message_classes)
        d.add(message_icon, time_stamp, p)
        shared_div.add_component(d, 0) 
        self.message.value = ''     # Clear message box after message is sent
        await shared_div.update()

def message_demo():
    wp = jp.WebPage()
    outer_div = jp.Div(classes='flex flex-col h-screen', a=wp)
    outer_div.add(header)
    d = jp.Div(classes='flex', a=outer_div)
    message = jp.Textarea(placeholder='Enter message here', a=d, classes=input_classes, debounce=500)
    send_button = jp.Button(a=d, click=send_message, classes=button_classes)
    send_button.add(button_icon, button_text)
    outer_div.add(shared_div)
    shared_div.add_page(wp)
    send_button.message = message
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("message_demo",message_demo, startup=message_initialize)



# Justpy Tutorial demo grid_test4 from docs/grids_tutorial/creating_grids.md
import justpy as jp

grid_options = """
{
    defaultColDef: {
        filter: true,
        sortable: true,
        resizable: true,
        cellStyle: {textAlign: 'center'},
        headerClass: 'font-bold'
    }, 
      columnDefs: [
      {headerName: "Make", field: "make"},
      {headerName: "Model", field: "model"},
      {headerName: "Price", field: "price"}
    ],
      rowData: [
      {make: "Toyota", model: "Celica", price: 35000},
      {make: "Ford", model: "Mondeo", price: 32000},
      {make: "Porsche", model: "Boxter", price: 72000}
    ]
}
"""

def grid_test4():
    wp = jp.WebPage()
    grid = jp.AgGrid(a=wp, options=grid_options)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("grid_test4",grid_test4)



# Justpy Tutorial demo alert_test2 from docs/tutorial/custom_components.md
import justpy as jp

class MyAlert(jp.Div):

    def __init__(self, **kwargs):
        self.title_text = 'This is the title'
        self.body_text = 'This is the body'
        super().__init__(**kwargs)
        root = self
        c1 = jp.Div(classes='bg-teal-100 border-t-4 border-teal-500 rounded-b text-teal-900 px-4 py-3 shadow-md',
                    role='alert', a=root)
        c2 = jp.Div(classes='flex', a=c1)
        c3 = jp.Div(classes='py-1', a=c2)
        c4 = jp.Svg(classes='fill-current h-6 w-6 text-teal-500 mr-4', xmlns='http://www.w3.org/2000/svg',
                    viewBox='0 0 20 20', a=c3)
        c5 = jp.Path(
            d='M2.93 17.07A10 10 0 1 1 17.07 2.93 10 10 0 0 1 2.93 17.07zm12.73-1.41A8 8 0 1 0 4.34 4.34a8 8 0 0 0 11.32 11.32zM9 11V9h2v6H9v-4zm0-6h2v2H9V5z',
            a=c4)
        c6 = jp.Div(a=c2)
        c7 = jp.P(classes='font-bold text-lg', a=c6, text=self.title_text)
        c8 = jp.P(classes='text-sm', a=c6, text=self.body_text)
        self.title_p = c7
        self.body_p = c8

    def react(self, data):
        self.title_p.text = self.title_text
        self.body_p.text = self.body_text
        self.text = ''


greetings = ['Bonjour', 'Hola', 'Zdravstvuyte', 'Nǐn hǎo', 'Salve', 'Konnichiwa', 'Guten Tag', 'Olá', 'Anyoung haseyo',
             'Asalaam alaikum', 'Goddag', 'Shikamoo', 'Goedendag', 'Yassas', 'Dzień dobry', 'Selamat siang', 'Namaste',
             'Merhaba', 'Shalom', ' God dag']


def translate(self, msg):
    self.title_text = 'Hello'

def alert_test2():
    wp = jp.WebPage()
    for greeting in greetings:
        d = MyAlert(a=wp, classes='m-2 w-1/4', title_text='hello', body_text='How is everybody?')
        d.on('click', translate)
        d.title_text = greeting
    return wp


# initialize the demo
from  examples.basedemo import Demo
Demo ("alert_test2",alert_test2)



"""
Created on 2022-09-22

@author: wf
"""
import json
import os
import sys
import traceback
from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter
from examples.basedemo import Demo
from jpcore.demoapp import JustpyDemoApp
from jpcore.utilities import find_files

class Demostarter:
    """
    start justpy demos on different ports
    """

    def __init__(self, base_path:str=None,mode: str = None, debug: bool = False):
        """
        constructor

        Args:
            base_path(str): the base_path
            debug(bool): if True switch on debug mode
        """
        Demo.testmode = True
        self.debug = debug
        self.mode = mode
        self.script_dir = os.path.dirname(__file__)
        if base_path is None:
            base_path=os.path.dirname(self.script_dir)
        if base_path.endswith("/examples"):
            base_path=os.path.dirname(base_path)
        self.justpy_dir = f"{base_path}/examples"
        self.example_json_file=f"{self.justpy_dir}/examples.json"
        if self.debug:
            print(f"collecting examples from {self.justpy_dir}")
        pymodule_files = find_files(self.justpy_dir, ".py")
        self.demos = []
        self.demos_by_name={}
        self.demos_by_source_file={}
        self.servers = {}
        self.errors = {}
        for pymodule_file in pymodule_files:
            demo = JustpyDemoApp(examples_dir=self.justpy_dir,pymodule_file=pymodule_file,debug=debug)
            if demo.is_demo:
                self.demos.append(demo)
                self.demos_by_name[demo.name]=demo
                self.demos_by_source_file[demo.source_file]=demo
            else:
                if self.debug:
                    print(f"{pymodule_file} is not a demo")
        self.add_video_links()
        if self.debug:
            print(f"found {len(self.demos)} justpy demo python modules")
            
    def add_video_links(self):
        """
        add video links to the demos
        """
        with open(self.example_json_file) as json_file:
            example_json = json.load(json_file)
            if "examples" in example_json:
                for i,video_record in enumerate(example_json["examples"]):
                    name=video_record.get("name",None)
                    video_url=video_record.get("video_url",None)
                    if name and video_url:
                        if name in self.demos_by_source_file:
                            demo=self.demos_by_source_file[name]
                            demo.video_url=video_url
                        elif name in self.demos_by_name:
                            demo=self.demos_by_name[name]
                            demo.video_url=video_url
                    else:
                        raise Exception(f"name or url missing for example #{i}")
                    
            pass
        
            
    def as_list_of_dicts(self)->list:
        """
        convert me to a list of dicts for tabular display
        
        Returns:
            list: a list of records/dicts with information about the demos
        """
        lod=[]
        for i,demo in enumerate(self.demos):
            try_it_link=""
            demo_link=f"""<a href="/demo/{demo.name}" target="_blank">{demo.name}</a>"""   
            if demo.try_it_url:
                try_it_link=f"""<a href="{demo.try_it_url}" target="_blank">try {demo.name}</a>"""
            record={
                "#":i+1,
                "→": demo.example_source.img_link,
                "name": demo_link,
                "try it": try_it_link,
                "📹": "✅" if demo.video_url is not None else "❌",
                "source": demo.source_link,
                "status": demo.status
            }
            lod.append(record)
        return lod      

    def start(self,limit=None, use_gather: bool = False):
        """
        start the demos  optionally limitting the number of demos

        Args:
            limit(int): the maximum number of demos to start (default: None)
            userGather(bool): if True try using gather
        """
        import justpy as jp
        tasklist=[]
        jp.justpy(start_server=False)
        app=jp.app
        for i, demo in enumerate(self.demos):
            try:
                print(f"mounting {i+1:3}:{demo}  ...")
                if use_gather:
                    tasklist.append(demo.mount(app))
                else:
                    demo.mount(app)
            except Exception as ex:
                self.errors[i] = ex
                print(f"failed due to exception: {str(ex)}")
                if self.debug:
                    print(traceback.format_exc())
            if limit is not None and i + 1 >= limit:
                break
        server=jp.get_server()
        server.run()
        #if use_gather:
        #    demo_results = await asyncio.gather(*tasklist, return_exceptions=True)
        pass

    async def stop(self):
        """
        try stopping all servers
        """
        for server in list(self.servers.values()):
            if self.debug:
                print(f"stopping server at port {server.port} ...")
            await server.stop()
            self.servers.pop(server.port)

def main(argv=None):  # IGNORE:C0111
    """main program."""

    if argv is None:
        argv = sys.argv
    try:
        program_name = os.path.basename(sys.argv[0])
        parser = ArgumentParser(
            description="justpy demo module starter",
            formatter_class=RawDescriptionHelpFormatter,
        )
        parser.add_argument(
            "-d", "--debug", dest="debug", action="store_true", help="show debug info"
        )
        parser.add_argument(
            "-m",
            "--mode",
            default=None,
            help="the server start mode process/direct or None for os specific (default: %(default)s)",
        )
        script_dir = os.path.dirname(__file__)
        parser.add_argument(
            "-p",
            "--path",
            default=os.path.dirname(script_dir),
            help="path to the examples (default: %(default)s)",
        )
        parser.add_argument(
            "--limit",
            type=int,
            default=None,
            help="limit the number of demos (default: %(default)s)",
        )

        args = parser.parse_args(argv[1:])
        demostarter = Demostarter(base_path=args.path,mode=args.mode, debug=args.debug)
        if args.debug:
            for demo in demostarter.demos:
                print(demo)
        demostarter.start(limit=args.limit)
    except Exception as e:
        indent = len(program_name) * " "
        sys.stderr.write(program_name + ": " + repr(e) + "\n")
        sys.stderr.write(indent + "  for help use --help")
        if args.debug:
            print(traceback.format_exc())
        return 2


if __name__ == "__main__":
    sys.exit(main())



"""
Created on 2022-09-07

@author: wf
"""
import textwrap
from typing import List

from starlette.requests import Request


class Context:
    """
    legacy context handler, encapsulates context
    """

    def __init__(self, context_dict: dict):
        """
        constructor

        Args:
            context_dict(dict): a context dict in legacy format
        """
        self.context_dict = context_dict
        self.page_options = PageOptions(context_dict.get("page_options", {}))
        self.use_websockets_js=self.context_dict.get("use_websockets","true")
        self.page_id_js=context_dict["page_id"]
        self.component_engine_type = context_dict["frontend_engine_type"]
        self.frontend_engine_libs = context_dict["frontend_engine_libs"]
        self.title_js=self.get_js_option("title", "JustPy")
        self.redirect_js=self.get_js_option("redirect","")
        self.display_url_js=self.get_js_option("display_url","")
        justpy_dict_js=str(self.context_dict.get("justpy_dict","[]"))
        self.justpy_dict_js=justpy_dict_js.replace('</' + 'script>', '</" + "script>')

    def get_url_for(self, name: str, path: str = None):
        """
        get url for the given route name
        Args:
            name: name of the route
            path: path params
        """
        url = None
        request = self.context_dict.get("request")
        if request is not None and isinstance(request, Request):
            if path is None:
                path = ""
            url = request.url_for(name, path=path)
        return url
    
    def get_js_option(self,key,default_value):
        """
        get the page_option with the given key as javascript using the
        default value in case the value is not set or none
        """
        js_option=self.page_options.page_options_dict.get(key,default_value)
        if js_option is None:
            js_option=default_value
        return js_option
        
    def as_html_lines(self,indent:str="  "):
        """
        generate the html lines for justpy to work 
        """
        html=self.as_script_src("justpy_core")
        html+=f"""{indent}<script>
{indent}  var page_id = {self.page_id_js};
{indent}  var use_websockets = {self.use_websockets_js};
{indent}  var justpyComponents = {self.justpy_dict_js};
"""
        html+=self.as_javascript_constructor(indent+"  ")
        html+=f"\n{indent}</script>\n{self.as_script_srcs(indent)}"
        html+=f"{indent}<script>\n{self.as_javascript_setup(indent)}\n{indent}</script>\n"
        return html
    
    def as_script_src(self,file_name:str, indent:str="  ", subdir=""):
        src= f"{indent}<script src='/templates/js/{subdir}/{file_name}.js'></script>\n"
        return src
    
    def as_script_srcs(self,indent:str="  "):
        """
        generate a list of javascript files to be imported
        """
        srcs = ""
        srcs +=self.as_script_src("event_handler", indent)
        for file_name in self.frontend_engine_libs:
            srcs +=self.as_script_src(file_name, indent, subdir=self.component_engine_type)
        return srcs
    
    def as_javascript_setup(self,indent):
        """
        generate the java script setup code
        """
        js=f"{indent}  justpy_core.setup();"
        return js

    def as_javascript_constructor(self,indent:str="    "):
        """
        generate my initial JavaScript
        """
        debug = str(self.page_options.get_debug()).lower()
        page_ready = str(self.page_options.get_page_ready()).lower()
        result_ready = str(self.page_options.get_result_ready()).lower()
        reload_interval_ms = self.page_options.get_reload_interval_ms()
        static_resources_url = self.get_url_for("static")
        if static_resources_url is None:
            static_resources_url = "/"
        javascript = f"""let justpy_core=new JustpyCore(
            this, // window
            {self.page_id_js}, // page_id
            '{self.title_js}', // title
            '{self.use_websockets_js}', // use_websockets
            '{self.redirect_js}', // redirect
            '{self.display_url_js}', // display_url
            {page_ready}, // page_ready
            {result_ready}, // result_ready     
            {reload_interval_ms}, // reload_interval
            {self.page_options.events},  // events
            '{static_resources_url}',  // static_resources_url
            {debug}   // debug
        );"""
        javascript = textwrap.indent(javascript, indent)
        return javascript


class PageOptions:
    """
    legacy page_options handler, encapsulating page_options
    """

    def __init__(self, page_options_dict: dict):
        """
        constructor
        """
        self.page_options_dict = page_options_dict
        self.events: List[str] = page_options_dict.get("events", [])

    def get_debug(self):
        """
        Get debug status of the PageOptions
        """
        return self.page_options_dict.get("debug", False)

    def get_page_ready(self):
        return "page_ready" in self.events

    def get_result_ready(self):
        return "result_ready" in self.events

    def get_reload_interval_ms(self) -> float:
        reload_interval = self.page_options_dict.get("reload_interval", 0)
        if reload_interval:
            ms = round(reload_interval * 1000)
        else:
            ms = 0
        return ms



# Justpy Tutorial demo quasar_example1 from docs/quasar_tutorial/introduction.md
import justpy as jp
import random

async def my_click(self, msg):
    self.color = random.choice(['primary', 'secondary', 'accent', 'dark', 'positive',
                                'negative','info', 'warning'])
    self.label = self.color
    msg.page.dark = not msg.page.dark
    await msg.page.set_dark_mode(msg.page.dark)

def quasar_example1():
    wp = jp.QuasarPage(dark=True)  # Load page in dark mode
    d = jp.Div(classes='q-pa-md q-gutter-sm', a=wp)
    jp.QBtn(color='primary', icon='mail', label='On Left', a=d, click=my_click)
    jp.QBtn(color='secondary', icon_right='mail', label='On Right', a=d, click=my_click)
    jp.QBtn(color='red', icon='mail', icon_right='send', label='On Left and Right', a=d, click=my_click)
    jp.Br(a=d)
    jp.QBtn(icon='phone', label='Stacked', stack=True, glossy=True, color='purple', a=d, click=my_click)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("quasar_example1",quasar_example1)



"""
Created on 2022-09-14

@author: th
"""
import unittest
from starlette.requests import Request

import justpy as jp
from jpcore.justpy_app import JustpyApp
from jpcore.justpy_config import SESSION_COOKIE_NAME
from tests.base_client_test import BaseClienttest


class TestJustpyApp(BaseClienttest):
    """
    Tests for JustpyApp features
    """

    @unittest.skipIf(BaseClienttest.inPublicCI(), "Setting of cookies currently does not work in the middleware")
    def test_issue_531(self):
        """
        tests error in session handling on invalid session id
        https://github.com/justpy-org/justpy/issues/531
        """
        @jp.app.middleware("http")
        async def add_process_time_header(request: Request, call_next):
            cookies = request.cookies
            cookies[SESSION_COOKIE_NAME] = "invalid token"
            # func = request.app.get_func_for_request(request)
            response = await call_next(request)
            return response

        @jp.app.route("/webpage", name="webpage")
        @jp.app.response
        def hello_world(_request):
            wp = jp.WebPage()
            _d = jp.Div(text='I am justpy webpage', a=wp)
            return wp

        #self.app.prioritize_routes()
        _response=self.checkResponse("/webpage",debug=False)
        self.assertIn("Bad Session", _response.text)


# https://github.com/elimintz/justpy/issues/387
# https://justpy.io/tutorial/custom_components/#tab-group-component
import justpy as jp
from examples.tutorial.tab_group_component import Tabs, TabsPills
import pandas as pd


def tab_comp_test():

    wp = jp.WebPage()

    dfs = [
        pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]}),
        pd.DataFrame({"bb": range(10), "cc": range(10)}),
        pd.DataFrame(
            {
                "foo": [1001, 1002, 1003, 1004],
                "bar": range(4),
                "blah": [1234, 2345, 3456, 4567],
            }
        ),
        pd.DataFrame({"c1": range(7), "c2": range(7)}),
    ]

    t = Tabs(a=wp, classes="w-3/4 m-4", style="", animation=False, content_height=550)
    for idx, chart_type in enumerate(["bar", "column", "line", "spline"]):
        d = jp.Div(style=Tabs.wrapper_style, delete_flag=True)
        l = jp.Label(a=d, text=chart_type)
        dfs[idx].jp.ag_grid(a=d, style="width: 100%")
        t.add_tab(f"id{chart_type}", f"{chart_type}", d)

    return wp


from examples.basedemo import Demo

Demo("Issue 387 aggrid tabgroup", tab_comp_test)



# Justpy Tutorial demo quasar_expansion_item2 from docs/quasar_tutorial/quasar_components.md
import justpy as jp

class My_expansion(jp.QExpansionItem):

    image_num = 10      # Start from image 10, previous are boring to my taste

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        outer_div = jp.Div(classes="q-pa-md", a=self)
        jp.QImg(src=f'https://picsum.photos/400/300/?image={My_expansion.image_num}', a=outer_div)
        self.icon = "photo"
        self.label = f'Image {My_expansion.image_num}'
        self.value = True
        self.expand_separator = True
        self.header_class = "bg-teal text-white text-overline"
        My_expansion.image_num += 1


def quasar_expansion_item2(request):
    wp = jp.QuasarPage(dark=False)
    d = jp.Div(classes="q-pa-md ", style="max-width: 500px", a=wp)
    jp.Link(href='https://quasar.dev/vue-components/expansion-item', text='Quasar Expansion Item Example', target='_blank',
            classes="text-h5 q-mb-md", a=d, style='display: block;')
    add_btn = jp.QBtn(label='Add Image', classes="q-mb-md", color='primary', a=d)
    close_btn = jp.QBtn(label='Close All', classes="q-ml-md q-mb-md", color='negative', a=d)
    open_btn = jp.QBtn(label='Open All', classes="q-ml-md q-mb-md", color='positive', a=d)
    l = jp.QList(bordered=True, a=d)
    wp.list = l
    l.add_component(My_expansion(), 0)

    def add_pic(self, msg):
        msg.page.list.add_component(My_expansion(), 0)
    add_btn.on('click', add_pic)

    def close_pics(self, msg):
        for c in msg.page.list.components:
            c.value = False
    close_btn.on('click', close_pics)

    def open_pics(self, msg):
        for c in msg.page.list.components:
            c.value = True
    open_btn.on('click', open_pics)

    return wp


# initialize the demo
from  examples.basedemo import Demo
Demo ("quasar_expansion_item2",quasar_expansion_item2)



"""
Created on 2022-08-27

@author: wf

see https://justpy.io/quasar_tutorial/introduction/
"""
import justpy as jp
import random


async def my_click(self, msg):
    self.color = random.choice(
        [
            "primary",
            "secondary",
            "accent",
            "dark",
            "positive",
            "negative",
            "info",
            "warning",
        ]
    )
    self.label = self.color
    msg.page.dark = not msg.page.dark
    await msg.page.set_dark_mode(msg.page.dark)


def quasar_example():
    wp = jp.QuasarPage(dark=True)  # Load page in dark mode
    d = jp.Div(classes="q-pa-md q-gutter-sm", a=wp)
    jp.QBtn(color="primary", icon="mail", label="On Left", a=d, click=my_click)
    jp.QBtn(color="secondary", icon_right="mail", label="On Right", a=d, click=my_click)
    jp.QBtn(
        color="red",
        icon="mail",
        icon_right="send",
        label="On Left and Right",
        a=d,
        click=my_click,
    )
    jp.Br(a=d)
    jp.QBtn(
        icon="phone",
        label="Stacked",
        stack=True,
        glossy=True,
        color="purple",
        a=d,
        click=my_click,
    )
    return wp


from examples.basedemo import Demo

Demo("quasar example", quasar_example)



"""
Created on 2022-09-07

@author: wf
"""
from tests.basetest import Basetest
from jpcore.template import Context

class TestTemplate(Basetest):
    """
    Tests template handling
    """
    
    def setUp(self, debug=False, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)
        self.context_dict = {
            "html": "",
            "justpy_dict": '[{"attrs": {}, "id": null, "vue_type": "html_component", '
            '"show": true, "events": [], "event_modifiers": {}, "classes": '
            '"text-5xl m-2", "style": "", "set_focus": false, "html_tag": '
            '"p", "class_name": "P", "event_propagation": true, '
            '"inner_html": "", "animation": false, "debug": false, '
            '"transition": null, "directives": {}, "scoped_slots": {}, '
            '"object_props": [], "text": "Hello there!"}]',
            "options": {
                "aggrid": True,
                "aggrid_enterprise": False,
                "bokeh": False,
                "component_file_list": [],
                "deckgl": False,
                "highcharts": True,
                "katex": False,
                "no_internet": True,
                "plotly": False,
                "quasar": False,
                "quasar_version": None,
                "static_name": "static",
                "tailwind": True,
                "vega": False,
            },
            "page_id": 0,
            "frontend_engine_type": "vue",
            "frontend_engine_libs": ['html_component', 'quasar_component', 'chartjp', 'aggrid', 'iframejp', 'deckgl', 'altairjp', 'plotlyjp', 'bokehjp', 'katexjp', 'editorjp', 'component_generator'],
            "page_options": {
                "body_classes": "",
                "body_html": "",
                "body_style": "",
                "css": "",
                "dark": False,
                "debug": False,
                "display_url": None,
                "events": [],
                "favicon": "",
                "head_html": "",
                "highcharts_theme": None,
                "redirect": None,
                "reload_interval": None,
                "title": "JustPy",
            },
            "request": None,
            "use_websockets": "true",
        }
        self.context_obj = Context(self.context_dict)

    def test_html(self):
        """
        test html generation
        """
        html=self.context_obj.as_html_lines()
        print(html)
        
    def test_javascript(self):
        """
        test javascript generation
        """
        js = self.context_obj.as_javascript_constructor()
        debug = True
        if debug:
            print(js)
        for param in [
            "window",
            "page_id",
            "title",
            "use_websockets",
            "redirect",
            "display_url",
            "page_ready",
            "result_ready",
            "reload_interval",
            "events",
            "static_resources_url",
            "debug",
        ]:
            self.assertIn(f"// {param}", js)



# Justpy Tutorial demo for_example3 from docs/blog/vue_comparison.md
import justpy as jp


def for_example3():
    wp = jp.WebPage(tailwind=False)
    ul = jp.Ul(a=wp)
    object = {
      'title': 'How to do lists in Vue',
      'author': 'Jane Doe',
      'publishedAt': '2016-04-10'
    }
    for name, value in object.items():
        jp.Li(text=f'{name}: {value}', a=ul)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("for_example3",for_example3)



"""
Created on 2022-09-01

@author: wf
"""
from selenium import webdriver
from typing import List
from sys import platform
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver import ChromeOptions
from selenium.webdriver import FirefoxOptions
from selenium.webdriver.chrome.service import Service as ChromeService
from webdriver_manager.chrome import ChromeDriverManager


class SeleniumBrowsers:
    """
    handle available selenium drivers/browsers
    """

    def __init__(self, headless: bool = True, cache_valid_range=7):
        self.headless = headless
        self.cache_valid_range = cache_valid_range
        self.browsers = {}

    def getFirst(self):
        """
        get the first available browser
        """
        if len(self.browsers.values()) == 0:
            self.getBrowsers()
        browserList = list(self.browsers.values())
        if len(browserList) > 0:
            return browserList[0]
        else:
            return None

    def getBrowsers(self) -> List[WebDriver]:
        """
        Returns a list of all browsers to test
        """
        if platform == "linux":
            self.browsers["firefox"] = self._getFirefoxWebDriver()
        else:
            self.browsers["chrome"] = self._getChromeWebDriver()
        return self.browsers

    def _getFirefoxWebDriver(self) -> webdriver.Firefox:
        """
        Returns the Firefox selenium webdriver
        """
        options = FirefoxOptions()
        options.headless = self.headless
        browser = webdriver.Firefox(options=options)
        return browser

    def _getChromeWebDriver(self) -> webdriver.Chrome:
        """
        Returns the Chrome selenium webdriver
        """
        options = ChromeOptions()
        options.headless = self.headless
        chrome_executable = ChromeDriverManager(cache_valid_range=365).install()
        service = ChromeService(chrome_executable)
        browser = webdriver.Chrome(service=service, options=options)
        return browser



# Justpy Tutorial demo commands_demo1 from docs/tutorial/working_with_html.md
import justpy as jp

def commands_demo1():
    wp = jp.WebPage()
    c = jp.parse_html("""
        <div>
        <p class="m-2 p-2 text-red-500 text-xl">Paragraph 1</p>
        <p class="m-2 p-2 text-blue-500 text-xl">Paragraph 2</p>
        <p class="m-2 p-2 text-green-500 text-xl">Paragraph 3</p>
        </div>
        """, a=wp)
    for i in c.commands:
        print(i)
        jp.Div(text=i, classes='font-mono ml-2', a=wp)
    print()
    c = jp.parse_html("""
            <div>
            <p class="m-2 p-2 text-red-500 text-xl">Paragraph 1</p>
            <p class="m-2 p-2 text-blue-500 text-xl">Paragraph 2</p>
            <p class="m-2 p-2 text-green-500 text-xl">Paragraph 3</p>
            </div>
            """, a=wp, command_prefix='justpy.')
    for i in c.commands:
        print(i)
        jp.Div(text=i, classes='font-mono ml-2', a=wp)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("commands_demo1",commands_demo1)



# Justpy Tutorial demo grid_test16 from docs/grids_tutorial/pandas.md
import justpy as jp
import pandas as pd

wm_df = pd.read_csv('https://elimintz.github.io/women_majors.csv').round(2)

def grid_test16():
    wp = jp.WebPage()
    grid = jp.AgGrid(a=wp)
    grid.load_pandas_frame(wm_df)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("grid_test16",grid_test16)



# Justpy Tutorial demo quasar_rating_test2 from docs/quasar_tutorial/quasar_components.md
import justpy as jp

def quasar_rating_test2():
    wp = jp.QuasarPage()
    wp.tailwind = True
    num_stars = 3
    r = jp.QRating(size='2em', max=num_stars, color='primary', classes='m-2 p-2', a=wp, value=2, debounce=0)
    for i in range(1,num_stars + 1,1):
        t = jp.QTooltip(text=f'{i} rating')
        r.add_scoped_slot(f'tip-{i}', t)
    return wp


# initialize the demo
from  examples.basedemo import Demo
Demo ("quasar_rating_test2",quasar_rating_test2)



# Justpy Tutorial demo calculator_test2 from docs/tutorial/custom_components.md
from justpy import Div, Input, Button, WebPage, justpy
import justpy as jp

class Calculator(Div):

    btn_classes = 'w-1/4 text-xl font-bold p-2 m-1 border bg-gray-200 hover:bg-gray-700 shadow'
    layout_text = [['7', '8', '9', '*'], ['4', '5', '6', '-'], ['1', '2', '3', '+'], ['C', '0', '.', '=']]

    def __init__(self, **kwargs):

        super().__init__(**kwargs)
        self.value = 0
        self.tape = Input(classes='block p-2 m-2  border text-right text-sm bg-gray-200', a=self, readonly=True, value=' ', style='width: 90%')
        self.result = Input(classes='block p-2 m-2 border text-2xl text-right', a=self, readonly=True, value='0', style='width: 90%')
        for line in type(self).layout_text:
            d = Div(classes='flex w-auto m-2', a=self)
            for b in line:
                b1 = Button(text=b, a=d, classes=type(self).btn_classes, click=self.calculator_click)
                b1.calc = self

    @staticmethod
    async def calculator_click(self, msg):
        calc = self.calc
        try:
            tape_value = eval(calc.tape.value)
        except:
            tape_value = 0
        changed = False
        if self.text == 'C':
            calc.result.value = '0'
            calc.tape.value = ' '
            if calc.value != 0:
                calc.value = 0
                changed = True
        elif self.text == '=':
            if calc.value != tape_value:
                calc.value = tape_value
                changed = True
            calc.result.value = str(tape_value)
            calc.tape.value = str(tape_value)
        else:
            if calc.tape.value[-1] in '*+-/' or self.text in '*+-/':
                calc.tape.value += ' ' + self.text
            else:
                calc.tape.value += self.text
            try:
                tape_value = eval(calc.tape.value)
                calc.result.value = str(tape_value)
                if calc.value != tape_value:
                    calc.value = tape_value
                    changed = True
            except:
                pass
        if changed:
            if calc.has_event_function('change'):
                calc_msg = msg
                calc_msg.event_type = 'change'
                calc_msg.id = calc.id
                calc_msg.button_text = self.text
                calc_msg.value = calc.value
                calc_msg.class_name = calc.__class__.__name__
                return await calc.run_event_function('change', calc_msg)


def calc_change(self, msg):
    print('In change')
    print(msg)
    self.d.text = self.value

def calculator_test2():
    wp = WebPage()
    c = Calculator(a=wp, classes='m-1 border inline-block', style='width: 250px', change=calc_change)
    d = Div(classes='border m-2 p-1 w-64 text-xl', text='0', a=wp)
    c.d = d
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("calculator_test2",calculator_test2)



# Justpy Tutorial demo bokeh_test from docs/tutorial/bokeh.md
import justpy as jp
import bokeh
bokeh.sampledata.download()  # Run this the first time you run the program
from bokeh.plotting import figure, output_file, show
from bokeh.sampledata.iris import flowers
from bokeh.embed.standalone import json_item
from bokeh.models import LogColorMapper
from bokeh.palettes import Viridis6 as palette1
from bokeh.plotting import figure
from bokeh.sampledata.unemployment import data as unemployment
from bokeh.sampledata.us_counties import data as counties1



def create_iris_figure(*, width=500, height=500):
    colormap = {'setosa': 'red', 'versicolor': 'green', 'virginica': 'blue'}
    colors = [colormap[x] for x in flowers['species']]

    p = figure(title = "Iris Morphology", plot_width=width, plot_height=height)
    p.xaxis.axis_label = 'Petal Length'
    p.yaxis.axis_label = 'Petal Width'

    p.circle(flowers["petal_length"], flowers["petal_width"],
             color=colors, fill_alpha=0.2, size=10)
    return p


def create_texas_figure(*, width=500, height=500):
    palette = tuple(reversed(palette1))

    counties = {
        code: county for code, county in counties1.items() if county["state"] == "tx"
    }

    county_xs = [county["lons"] for county in counties.values()]
    county_ys = [county["lats"] for county in counties.values()]

    county_names = [county['name'] for county in counties.values()]
    county_rates = [unemployment[county_id] for county_id in counties]
    color_mapper = LogColorMapper(palette=palette)

    data = dict(
        x=county_xs,
        y=county_ys,
        name=county_names,
        rate=county_rates,
    )

    TOOLS = "pan,wheel_zoom,reset,hover,save"

    p = figure(
        title="Texas Unemployment, 2009", tools=TOOLS,
        x_axis_location=None, y_axis_location=None,
        tooltips=[
            ("Name", "@name"), ("Unemployment rate", "@rate%"), ("(Long, Lat)", "($x, $y)")
        ])
    p.grid.grid_line_color = None
    p.hover.point_policy = "follow_mouse"

    p.patches('x', 'y', source=data,
              fill_color={'field': 'rate', 'transform': color_mapper},
              fill_alpha=0.7, line_color="white", line_width=0.5)
    return p



def bokeh_test(request):
    wp = jp.WebPage(tailwind=True)
    p = create_iris_figure()
    wp.c = jp.BokehChart(chart=p, a=wp)
    p1 = create_iris_figure(width=300, height=300)
    jp.BokehChart(chart=p1, a=wp)
    p2 = create_texas_figure()
    jp.BokehChart(chart=p2, a=wp)
    return wp


# initialize the demo
from  examples.basedemo import Demo
Demo ("bokeh_test",bokeh_test, BOKEH=True)



# Justpy Tutorial demo transition_test from docs/tutorial/transitions.md
import justpy as jp

def toggle_hidden(self, msg):
    for d in self.div_list:
        if d.has_class('hidden'):
            d.remove_class('hidden')
        else:
            d.set_class('hidden')


def transition_test():
    wp = jp.WebPage()
    tran = jp.Dict()
    tran.load = "transition ease-out duration-1000"
    tran.load_start = 'opacity-0 transform scale-0'
    tran.load_end = 'opacity-100 transform scale-100'
    tran.enter = 'transition ease-out duration-1000'
    tran.enter_start = 'opacity-0 transform scale-0'
    tran.enter_end = 'opacity-100 transform scale-1000'
    tran.leave = "transition ease-out duration-1000"
    tran.leave_start = 'opacity-100 transform scale-1000'
    tran.leave_end = 'opacity-0 transform scale-0'
    div_list = []
    flex_div = jp.Div(classes='flex flex-wrap', a=wp)
    for i in range(10):
        d = jp.Div(a=flex_div, text='hello', style='height: 150px; width:200px')
        jp.Pre(text=' test', a=d)
        d.transition = tran
        d.classes = 'border text-xl m-2 h-16 w-16 bg-blue-500'
        div_list.append(d)
    b = jp.Button(text='Click here', classes=jp.Styles.button_simple + ' m-2', a=wp, click=toggle_hidden)
    b.div_list = div_list
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("transition_test",transition_test)



"""
Created on 2022-09-18

@author: th
"""
import asyncio
import unittest
import pandas as pd
from selenium.webdriver import ActionChains
from selenium.webdriver.common.by import By

import justpy as jp
from tests.base_selenium_test import BaseSeleniumTest
from selenium.webdriver.support import expected_conditions as EC

class TestIssue304(BaseSeleniumTest):
    """
    testing issue 304
    https://github.com/justpy-org/justpy/issues/304
    """

    async def tearDown(self):
        """
        close selenium browser and stop server
        """
        self.browser.close()
        await self.server.stop()

    async def test_issue_30(self):
        """
        test when the issue occurs that input and input display differ
        """
        await self.server.start(self.issue_page)
        url = self.server.get_url("/")
        self.browser.get(url)
        timeout = 5
        driver = self.get_waiting_browser(self.browser)
        # wait until a row is clickable
        self.assertTrue(
                driver.until(
                        EC.element_to_be_clickable((By.CLASS_NAME, "ag-cell"))
                ),
                "After loading the page an ag-cell should be clickable"
        )
        # click cells since the row div does not register a click
        cells = self.browser.find_elements(By.CLASS_NAME, "ag-cell")
        cells[18*2].click()
        # test row single click
        self.assertTrue(
                driver.until(
                        EC.text_to_be_present_in_element(
                                locator=(By.NAME, "test_row_clicked"),
                                text_="Clicked")
                )
        )
        # test row double click
        action_chains = ActionChains(self.browser)
        action_chains.double_click(cells[18*3]).perform()
        self.assertTrue(
                driver.until(
                        EC.text_to_be_present_in_element(
                                locator=(By.NAME, "test_row_double_clicked"),
                                text_="Clicked")
                )
        )

    @unittest.skipIf(True, "Only for manual testing")
    def test_manual(self):
        """
        start the server to test the issue manually
        """
        jp.justpy(self.issue_page, port=8500)

    @staticmethod
    def get_sample_data() -> pd.DataFrame:
        """
        returns a dataframe with sample data
        """
        wm_df = pd.read_csv('https://elimintz.github.io/women_majors.csv').round(2)
        return wm_df

    @staticmethod
    def row_clicked(self, msg):
        """
        row select event handler
        """
        # print("row clicked (single)")
        if msg.selected:
            self.row_data_div.text = msg.data
            self.row_selected = msg.rowIndex
            self.a.test_row_clicked.text = "Clicked"
        elif self.row_selected == msg.rowIndex:
            self.row_data_div.text = ''

    @staticmethod
    def row_double_clicked(self, msg):
        """
        row select event handler
        """
        # print("row double-clicked")
        if msg.selected:
            self.row_data_div.text = msg.data
            self.row_selected = msg.rowIndex
            self.a.test_row_double_clicked.text = "Clicked"
        elif self.row_selected == msg.rowIndex:
            self.row_data_div.text = ''

    @staticmethod
    def issue_page():
        """
        page containing the issue
        """
        wp = jp.WebPage()
        row_data_div = jp.Div(a=wp)
        grid = TestIssue304.get_sample_data().jp.ag_grid(a=wp)
        grid.row_data_div = row_data_div
        grid.on('rowClicked', TestIssue304.row_clicked)
        grid.on('rowDoubleClicked', TestIssue304.row_double_clicked)
        grid.options.columnDefs[0].checkboxSelection = True
        grid.options.rowSelection = 'single'
        wp.test_row_clicked = jp.Div(a=wp, name="test_row_clicked")
        wp.test_row_double_clicked = jp.Div(a=wp, name="test_row_double_clicked")
        return wp



# Justpy Tutorial demo ripple_test from docs/quasar_tutorial/introduction.md
import justpy as jp
# https://quasar.dev/vue-directives/material-ripple#Ripple-API

def ripple_test():
    wp = jp.QuasarPage()
    d = jp.QDiv(classes="q-pa-md q-gutter-md row justify-center", a=wp)
    d1 = jp.QDiv(v_ripple={'center': True, 'color': 'orange-5'}, classes="relative-position container bg-grey-3 text-black inline flex flex-center", text='center',
                 style='border-radius: 50%; cursor: pointer; width: 150px; height: 150px', a=d)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("ripple_test",ripple_test)



"""
Created on 2022-09-01

@author: wf
"""
import aiohttp
import asynctest
from jpcore.justpy_app import JustpyServer
from tests.basetest import Basetest, Profiler


class BaseAsynctest(asynctest.TestCase):
    """
    basic asynch test
    """

    # https://github.com/encode/starlette/blob/master/docs/testclient.md

    async def setUp(
        self,
        port: int = 8123,
        host: str = "127.0.0.1",
        sleep_time=None,
        with_server=True,
        debug=False,
        profile=True,
        mode=None,
    ):
        """Bring server up.

        Args:
            wpfunc: the (async) function for the webpage
            port(int): the port
            host(str): the host
            sleep_time(float): the time to sleep after server process was started
            debug(bool): if True debugging is on
            profile(bool): if True time for test is profiled
        """
        if sleep_time is None:
            sleep_time = 2.0 if Basetest.inPublicCI() else 0.75
        self.sleep_time = sleep_time
        self.server = None
        if with_server:
            self.server = JustpyServer(
                port=port, host=host, sleep_time=sleep_time, mode=mode, debug=debug
            )
        self.debug = debug
        self.profile = profile
        msg = f"test {self._testMethodName}, debug={self.debug}"
        self.profiler = Profiler(msg, profile=self.profile)

    async def tearDown(self):
        """Shutdown the app."""
        if self.server is not None:
            await self.server.stop()
        self.profiler.time()

    def getUrl(self, path):
        """
        get the url for the given path

        Args:
            path(str): the path
        Returns:
            str: the url for the path

        """
        url = f"http://{self.server.host}:{self.server.port}{path}"
        return url

    async def getResponseHtml(self, path: str = "/"):
        """
        get the status and raw html for the given path

        Args:
            path(str): the path
        Returns:
            status,rawhtml
        """
        async with aiohttp.ClientSession() as session:
            url = self.getUrl(path)
            async with session.get(url) as resp:
                rawhtml = await resp.content.read()
                status = resp.status
        return status, rawhtml
